# -*- coding: utf-8 -*-

#Copyright (C) 2008 David Molina, Adrian Bravo Navarro

#This file is part of HDLorean.
#
#HDLorean is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#HDLorean is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with HDLorean.  If not, see <http://www.gnu.org/licenses/>.


from DBExceptions import *
from pysqlite2 import dbapi2 as sqlite
import os
import thread
import operator
import time
import random


##@brief
#
#PySquilte interface class
#
#This class implements all the functions needed 
#to control the journal database of HdLorean using pysqlite 2.0
class DataBaseWrapper:
	
	__sqlite3_cur = None
	__active = None
	__last = None

	##@brief
	#
	#The constructor of the class(it is a empty constructor).	
	def __init__(self):
		self.__active = {}


	##@brief	 
	# 
	#Create a connection to the HdLorean database, 
	#if the database does not exist create a new one.
	#
	#@param database String that contains the name of the Database to connect.
	def openCon(self,database):
		hdlorean_dir = os.getenv("HOME")+"/.hdlorean"
		if not os.path.exists(hdlorean_dir):
			os.mkdir(hdlorean_dir)
		tpid = self.getPid()
		if self.__active.has_key(tpid):
			self.__active[tpid]['cur'] = self.__active[tpid]['con'].cursor()
			self.__active[tpid]['links'] = self.__active[tpid]['links'] + 1
		else:
			conn =  sqlite.connect(hdlorean_dir+database+'.db',isolation_level=None)
			self.__active[tpid] = {'con':conn,'cur':conn.cursor(),'loop':0,'links':0}
		
		
	   
	##@brief 
	#
	#Create the table needed to the HdLorean aplication.
	#
	#@param table String that contains the create table query
	#@param index String that contains the create index query.
	# 
	#@throw SQLError if the quey is wrong.	 
	#@throw ConnError if there is not connection to the database.	 
	def createTable(self,table,index=None):
		tpid =self.getPid()
		try:
			self.__active[tpid]['cur'].execute(table)
			if index:
				self.__active[tpid]['cur'].execute(index)
		except sqlite.DatabaseError, e:
			raise SQLError, e
		except AttributeError, e:
			raise ConnError, e
	
		
	##@brief 
	#
	#Close connection to the database. 
	def closeCon(self):
		tpid = self.getPid()
		if not self.__active[tpid]['links']:
			self.__active[tpid]['cur'].close()
			self.__active[tpid]['con'].close()
			del self.__active[tpid]
		else:
			self.__active[tpid]['links'] = self.__active[tpid]['links'] - 1
	

	## @brief Get the identifier of the running process/thread
	#
	# @return Integer, the identifier of the process
	# @author Adrián, David
	def getPid(self):
		return thread.get_ident()


	## @brief Waits a random number of ms to avoid two querys colliding at the database
	#
	# @author Adrián	
	def __backoff(self,tpid):
		# Increase the number of loops
		loops = self.__active[tpid]['loop'] + 1
		self.__active[tpid]['loop'] = loops
		random.seed(tpid)
		if loops == 1:
			high = 31
		else:
			high = 2**loops
		r = operator.mod(random.randint(1,high),2000)
		# Wait for random in 1 to 2^loop (truncated) miliseconds before trying again
		c = operator.mod(operator.truediv(r,1000),2)
		print tpid, " he esperado ",self.__active[tpid]['loop'], "veces y ahora toca ",c
		time.sleep(c)
	

	
	## @brief Execute a query that returns a result. Use to execute select querys.
	#
	#@param query String with the query to execute.
	#@return An array with the selected tuples of the database.
	#
	#@throw SQLError if the query is wrong.
	#@throw ConnError if there is not connection to the database.   
	#
	# @author Adrián
	def resultQuery(self,query, argv=None):
		tpid = self.getPid()
		success = False
		while not success:
			try:
				if argv:
					self.__active[tpid]['cur'].execute(query,argv)
				else:
					self.__active[tpid]['cur'].execute(query)
				success = True
				self.__last = query
				result = [] #< keeps the result array
				for row in self.__active[tpid]['cur']:
					result.append(row)
				if result==[(None,)] or result==[]:
					return None
				else:
					return result 
			except sqlite.OperationalError,e:
				self.__backoff(tpid)
			except sqlite.DatabaseError, e:
				raise SQLError, e  
			except AttributeError, e:
				raise ConnError, e	


	##@brief Execute a query that has not result. Use to execute update, insert and delete querys.
	#   
	#@param query String with the query to execute.
	#
	#@throw IntegrityError if there is an integrity problem in the database.
	#@throw SQLError if the query is wrong.
	#@throw ConnError if there is not connection to the database.
	#
	# @author Jorge, Adrián
	def noResultQuery(self,query,argv=None):
		tpid = self.getPid()
		insertado = False
		while not insertado:
			try:
				if argv:
					self.__active[tpid]['cur'].execute(query,argv)
				else:
					self.__active[tpid]['cur'].execute(query)
				insertado = True
				self.__last = query
			except sqlite.OperationalError, e:
				self.__backoff(tpid)
			except sqlite.IntegrityError, e:
				raise IntegrityError, e
			except sqlite.DatabaseError, e:
				raise SQLError, e
			except AttributeError, e:
				raise ConnError, e
