# -*- coding: utf-8 -*-

"""
Copyright (C) 2008 David Molina, Salvador de la Puente, Federico G. Mon Trotti

This file is part of HDLorean.

HDLorean is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HDLorean is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HDLorean.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
from DataBaseWrapper import DataBaseWrapper, ConnError, SQLError, IntegrityError
from hdlogger import HDLogger


##@brief History database class implements all the functions needed 
# to control the History database of HdLorean.
#
class History:
	"""
	History database class implements all the functions needed 
	to control the History database of HdLorean.
	"""
	
	##@brief Contains the DataBaseWrapper.
	#
	__DBW = None
	
	##@brief Contains the logger.
	#
	__logger = None 
	
	##@brief Lets write logs messages.
	#
	__log = None
	
	##@brief Valid fileds of database.
	#
	__keys = None
	
	##@brief Default constructor.
	#
	def __init__(self):
		self.__DBW =  DataBaseWrapper()
		self.__logger = HDLogger.HDLogger()		
		self.__keys = None
		self.__log = self.__logger.setupHDLogger("hdlorean.db.History")

	##@brief Create tables and triggers needed by History database.
	#
	def createTables(self):
		"""
		Create tables and triggers needed by History database.
		"""
		try:
			self.__DBW.openCon('/HdHistory')
			try: #CREATE TIMELINE
				table = 'CREATE TABLE IF NOT EXISTS timeline (' 
				table = table + 'id INTEGER PRIMARY KEY AUTOINCREMENT, '
				table = table + 'timestamp TIMESTAMP NOT NULL, '
				table = table + 'initpath TEXT NOT NULL, '
				table = table + 'finalpath TEXT NOT NULL,'
				table = table + 'location VARCHAR(255) NOT NULL, '
				table = table + 'operation VARCHAR(255) NOT NULL, '
				table = table +	'requestor	VARCHAR(255) NOT NULL, '
				table = table + 'bsize INTEGER NOT NULL, '
				table = table +	'wrapper VARCHAR(255) NOT NULL); '
				index = 'CREATE INDEX IF NOT EXISTS indice ON timeline(timestamp) '
				self.__DBW.createTable(table, index)

				#CREATE TIME ATTRIBS
				table = 'CREATE TABLE IF NOT EXISTS time_attribs (' 
				table = table + 'timestamp TIMESTAMP PRIMARY KEY, '
				table = table + 'GMT INTEGER NOT NULL, '
				table = table + 'DST INTEGER NOT NULL); '
				self.__DBW.createTable(table)

				#CREATE FILE ATTRIBS
				table = 'CREATE TABLE IF NOT EXISTS file_attribs (' 
				table = table + 'id INTEGER PRIMARY KEY, '
				table = table + 'size INTEGER NOT NULL, '
				table = table + 'permissions INTEGER NOT NULL, '
				table = table + 'extended_permissions INTEGER NOT NULL, '
				table = table + 'mimetype VARCHAR(255) NOT NULL, '
				table = table + 'last_access TIMESTAMP NOT NULL, '	
				table = table + 'last_modification TIMESTAMP NOT NULL, '
				table = table + 'creation_time TIMESTAMP NOT NULL, '
				table = table + 'type VARCHAR(255) NOT NULL); '
				self.__DBW.createTable(table)
				
				#DEFINE DELETE ON CASCADE
				trigger = 'CREATE TRIGGER IF NOT EXISTS delete_all_info'
				trigger = trigger + '\nBEFORE DELETE ON timeline'
				trigger = trigger + '\nFOR EACH ROW BEGIN'
				trigger = trigger + '\nDELETE FROM file_attribs WHERE id = OLD.id;'
				trigger = trigger + '\nEND;'
				self.__DBW.noResultQuery(trigger)

				self.__log.info('History database created without problems.')   
				self.__keys = set([	'id', 'timestamp', 'operation', 'btype',
								 	'GMT', 'DST', 'status', 'size', 'location',
								 	'initpath',	'finalpath', 'requestor',
								 	'parent'
								 ])
			except SQLError, exc:
				self.__log.critical(
				'SQL error has happened in the database: %s, %s. Error: %s.',
				table, index, exc )
			except ConnError:
				self.__log.critical(
				'The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()	
	
	##@brief Inserts in the time_attribs table a new row with
	# the time attributes given.
	#
	#@param timeAttribs tuple with timestamp,GMT and DST.
	def insertTimeAttributes(self, timeAttribs):
		"""
		Inserts in the time_attribs table a new row with
		the time attributes given.
		"""
		try:
			self.__DBW.openCon('/HdHistory')
			try:
				insert = "insert into time_attribs (timestamp,GMT,DST) values (?,?,?)"
				self.__DBW.noResultQuery(insert, timeAttribs)
			except IntegrityError:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					insert, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
	##@brief Inserts in the file_attribs table a new row with the
	# file attributes given.
	#
	#@param fileAttribs tuple with
	# id, size, permissions, extended permissions, mime type,
	# acces time, modification time, creation time
	# and type (F for file, D for directory)
	def insertFileAttributes(self, fileAttribs):
		"""
		Inserts in the file_attribs table a new row with the
		file attributes given.
		"""
		try:
			self.__DBW.openCon('/HdHistory')
			try:
				insert = "insert into file_attribs "\
						+ "(id,size,permissions,extended_permissions,mimetype, "\
						+ "last_access,last_modification,creation_time,type) " \
						+ "values (?,?,?,?,?,?,?,?,?)"
				self.__DBW.noResultQuery(insert, fileAttribs)
			except IntegrityError:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					insert, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			

	##@brief Inserts in the timeline table a new row with the time
	# attributes given.
	#
	#@param versionInfo tuple with
	# timestamp, init_path, final_path, location,
	# operation, requestor, bsize and wrapper.
	#@return id of new the table entry
	def insertNewVersion(self, versionInfo):
		"""
		Inserts in the timeline table a new row with the time
		attributes given.
		"""
		try:
			identifier = 0
			self.__DBW.openCon('/HdHistory')
			try:
				query = 	  "insert into timeline "\
							+ "(timestamp, initpath, finalpath, location, "\
							+ "operation, requestor, bsize, wrapper) "\
							+ "values (?,?,?,?,?,?,?,?)"
							
				self.__DBW.noResultQuery(query, versionInfo)
				query = "select id from timeline where timestamp = ? and finalpath = ?"
				argv = (versionInfo[0], versionInfo[2])
				identifier = self.__DBW.resultQuery(query, argv)[0][0]
			except IntegrityError, exc:
				self.__log.critical(
					'Integrity error has happened in the database.Error: %s',
					exc)
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.', query, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			return identifier

	##@brief Update the old_timestamp to new_timestamps.
	#
	#@param path String with the path of the file.
	#@param old_timestamp The timestamp in the database
	#@param new_timestamp The new timestamp.
	def updateTimestamp(self, path, old_timestamp, new_timestamp):
		"""
		Update the old_timestamp to new_timestamps.
		"""
		try:
			self.__DBW.openCon('/HdHistory')
			try:
				update = 	  "update timeline set "\
							+ "timestamp=? where finalpath=? and timestamp=?;"
				argv = (new_timestamp, path, old_timestamp)
				self.__DBW.noResultQuery(update, argv)				
			except IntegrityError:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					update, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
	##@brief Gets the oldest timestamp in the data base.
	#
	#@return 0 if there is not result or a timestamp.
	def firstTimestamp(self):
		"""
		Gets the oldest timestamp in the data base.
		"""
		try:
			timeS = 0
			self.__DBW.openCon('/HdHistory')
			try:
				select = "select timestamp from timeline order by timestamp limit 1;"
				result = self.__DBW.resultQuery(select)
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
			if result != None:
				timeS = int(result[0][0])
			return timeS		 
			
	
	##@brief Gets the latest timestamp in the data base.
	#
	#@return 0 if there is not result or a timestamp.
	def lastTimestamp(self):
		"""
		Gets the latest timestamp in the data base.
		"""
		try:
			timeS = 0
			self.__DBW.openCon('/HdHistory')
			try:
				select = "select timestamp from timeline order by timestamp DESC limit 1;"
				result = self.__DBW.resultQuery(select)
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.', 
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
			if result != None:
				timeS = int(result[0][0])
			return timeS	
		
	
	##@brief Search for the next timestamp of the param.
	#
	#@param timestamp reference timestamp
	#@return 0 if there isnt result or next timestamp.
	def prevTimestamp(self, timestamp):
		"""
		Search for the next timestamp of the param.
		"""
		try:
			result = None
			self.__DBW.openCon('/HdHistory')
			try:
				select =  "select timestamp from timeline where timestamp<? "\
						+ "order by timestamp DESC limit 1;"
				argv = (str(timestamp),)
				result = self.__DBW.resultQuery(select, argv)
				timeS = 0
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.', 
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
			if result != None:
				timeS = int(result[0][0])
			return timeS		 
	

	##@brief Gets timestamps in an interval.
	#
	#@param initTimestamp reference initial timestamp
	#@param finalTimestamp reference final timestamp
	#@return timestamps in an interval.
	def getTimestampsBetween(self, initTimestamp, finalTimestamp):
		"""
		Gets timestamps in an interval.

		"""
		try:
			timeList = []
			self.__DBW.openCon('/HdHistory')
			try:
				select =  "select timestamp from timeline where timestamp >= ? "\
						+ "and timestamp <= ? order by timestamp ASC;"
				argv = (initTimestamp, finalTimestamp)
				result = self.__DBW.resultQuery(select, argv)
				timeList = []
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()  
			
			if result and result != [()]:
				for actualTime in result:
					result.append(actualTime[0])
			return timeList	 

	##@brief Search for the previous timestamp of the param.
	#
	#@param timestamp reference timestamp
	#@return None if there isnt result or the previous timestamp.
	def nextTimestamp(self, timestamp):
		"""
		Search for the previous timestamp of the param.
		"""
		try:
			timeS = 0
			self.__DBW.openCon('/HdHistory')
			try:
				select =  "select timestamp from timeline where timestamp>? "\
						+ "order by timestamp ASC limit 1;"
				argv = (str(timestamp),)
				result = self.__DBW.resultQuery(select, argv)
				if result != None:
					timeS = int(result[0][0])
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()  		
			return timeS
		
	##@brief Search for the wrapper that made a backup in the timestamp given.
	#
	#@param timestamp unix timestamp to check its wrapper.
	#@return codename of the wrapper used at that timestamp or "" if error.
	def whichWrapper(self, timestamp):
		"""
		Search for the wrapper that made a backup in the timestamp given.
		"""
		try:
			result = None
			self.__DBW.openCon('/HdHistory')
			try:
				select = "select wrapper from timeline where timestamp=?;"
				argv = (timestamp,)
				result = self.__DBW.resultQuery(select, argv)
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()			
			
			if result != None:
				wrapper = result[0]
				return wrapper[0]
			else:
				return ""			
		
	##@brief Search for the wrapper that made backups for a path in an interval.
	#
	#@param initTimestamp low timestamp of the interval. By default, it's 0.
	#@param finalTimestamp hight timestamp of the interval. By default, it's 0.
	#@return a list with codenames of the wrappers used inside that interval.
	def getWrappersBetween(self, initTimestamp, finalTimestamp, path):
		"""
		Search for the wrapper that made backups for a file in an interval.
		"""
		try:
			result = None
			listArgs = []
			self.__DBW.openCon('/HdHistory')
			try:
				slashpath = path+"/%%"
				select =  "select distinct wrapper from timeline where "\
						+ "? <= timestamp and timestamp <= ? and "\
					 	+ "(finalpath like ? or finalpath like ?)"
				listArgs.extend([initTimestamp, finalTimestamp, slashpath, path])
				result = self.__DBW.resultQuery(select, tuple(listArgs))
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			return result

	##@brief Delete all versions of a path in an interval.
	#
	#@param path path of version to delete
	#@param initTimestamp low timestamp of the interval.
	#@param finalTimestamp hight timestamp of the interval.
	def deleteVersions(self, path, initTimestamp, finalTimestamp):
		"""
		Delete all versions of a path in an interval.
		"""
		if os.path.normpath(path) == "/":
			path = ""			
					
		try:				
			self.__DBW.openCon('/HdHistory')
			try:
				delete =  "delete from timeline where "\
						+ "(finalpath = ? or finalpath like ?) and "\
						+ "(timestamp >= ? and timestamp <= ?);"
				slashPath = path + "/%%"
				argv = (path, slashPath, initTimestamp, finalTimestamp)
				print delete
				print argv
				self.__DBW.noResultQuery(delete, argv)
			except IntegrityError:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					delete, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()

	##@brief Gets info of the backups of the path.
	#
	#@param path String with a path.
	#
	#@return An array with the number of backups,the date of the last backup,
	#the size of the file. In that order.
	def backUpInfo(self, path, recursive = False):
		"""
		Gets info of the backups of the path.
		"""
		try:
			self.__DBW.openCon('/HdHistory')
			try:
				slashpath = path+"/%%"
				if recursive:
					select =   "select * from timeline where "\
							 + "(finalpath like ? or finalpath like ?) "\
							 + "order by timestamp DESC;"
					argv = (slashpath, path)
				else:
					select =   "select * from timeline where finalpath=? order "\
							 + "by timestamp DESC;"
					argv = (path)
				result = self.__DBW.resultQuery(select, argv)
				if (result != None):			
					versions = len(result)
					row = result[0]
					return versions, row[1], row[7]
				else:
					return None
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
	##@brief Gets basic info about all contents versioned once of a directory.
	#
	#@param path path to recover its contents
	#@param oneLevel if it is true then, method just recover sons of
	# one deep level.
	#@return a list with paths, type of each path and MIME Type.
	def getDeepDirList(self, path, oneLevel = False):
		"""
		Gets basic info about all contents versioned once of a directory.
		"""
		if os.path.normpath(path) == "/":
			path = ""			
		mostRecentTimestamp = self.lastTimestamp()
					
		try:
			result = None
			self.__DBW.openCon('/HdHistory')
			try:
				slashpath = path + "/%%"				
				innerSlashpath = slashpath + "/%%"
				argv = (mostRecentTimestamp, slashpath, slashpath)
				
				# Just one level
				restriction = ""
				if oneLevel:
					restriction = "and finalpath not like ?"
					argv = (mostRecentTimestamp, path, slashpath, innerSlashpath)
									
				select =  "select timeline.finalpath, "\
						+ "file_attribs.type, file_attribs.mimetype, "\
						+ "timeline.timestamp from timeline left join "\
						+ "file_attribs on timeline.id = file_attribs.id  "\
						+ "where timeline.timestamp <= ? and "\
						+ "finalpath <> ? and "\
						+ "finalpath like ? " + restriction + " order by "\
						+ "file_attribs.type ASC, timeline.finalpath ASC, "\
						+ "timeline.timestamp DESC;"
				
				result = self.__DBW.resultQuery(select, argv)
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
			if result and result != [()]:
				return result
			else:
				return []
			
	##@brief Gets the historical of versions for a path.
	#
	# The historical of a path is a list with all versions tracked
	# over the path and same info of getVersionInfo.
	#
	#@param path path of file or directory. If path is a directory then,
	# operations over sons of that directory are also shown inside historical.
	#@param timestamp by default is None, but given a unix timestamp, returns
	# only that version info.
	#
	#@return a pair with a list of tuples with the next information
	# (it can be only one if timestamp is specified):
	# ID
	# timestamp
	# init path
	# operation path
	# location
	# operation
	# requestor
	# backup size
	# name of sm
	# file size
	# permissions
	# x permissions
	# mime type
	# last access
	# last modif
	# creation time
	# type (F or D)
	#
	# And total number of backups too.
	def getVersionsInfo(self, path, timestamp = None):
		"""
		Gets the historical of versions for a path.
		"""
		if os.path.normpath(path) == "/":
			path = ""
		
		try:
			historical = []
			numberOfSnapshots = 0
			self.__DBW.openCon('/HdHistory')
			try:
				if timestamp:
					operation = "="
				else:
					timestamp = self.lastTimestamp()
					operation = "<="

				
				slashpath = path+"/%%"									
				select =  "select timeline.*, file_attribs.* from timeline "\
						+ "inner join file_attribs on "\
						+ "timeline.id = file_attribs.id "\
						+ "where timestamp " + operation + " ? and "\
						+ "(finalpath = ? or finalpath like ?) order by "\
						+ "timestamp DESC;"
				argv = (timestamp, path, slashpath)
				result = self.__DBW.resultQuery(select, argv)
				
				if result != [()] and result:
					# For each backup inside history
					numberOfSnapshots = len(result) 
					for actualBackup in result:
						historical.append((	actualBackup[0],	# ID
											actualBackup[1],	# timestamp
											actualBackup[2],	# init path
											actualBackup[3],	# operation path
											actualBackup[4],	# location
											actualBackup[5],	# operation
											actualBackup[6],	# requestor
											actualBackup[7],	# backup size
											actualBackup[8],	# name of sm
											actualBackup[10],	# file size
											actualBackup[11],	# permissions
											actualBackup[12],	# x permissions
											actualBackup[13],	# mime type
											actualBackup[14],	# last access
											actualBackup[15],	# last modif
											actualBackup[16],	# creation time
											actualBackup[17],	# type (F or D)
										 ))
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
			return historical, numberOfSnapshots
		
	##@brief Gets a quick briefing of all versions of a path
	#
	# The briefing some quick info about the versions of a path
	#
	#@param path path of file or directory. If path is a directory then,
	# operations over sons of that directory are also shown inside briefing.
	#@param timestamp by default is None, but given a unix timestamp, returns
	# only that version info.
	#
	#@return a pair with a list of tuples with the next information
	# timestamp
	# operation path
	# operation
	# backup size
	# file size
	# amount of backed up data*
	#
	# And total number of backups too.
	#
	# If version has more than 1000 files, -1 is returned
	# TODO: Extends note with average of data size and composition
	def getVersionsBriefing(self, path, timestamp = None):
		"""
		Gets a quick briefing of all versions of a path
		"""
		if os.path.normpath(path) == "/":
			path = ""
		
		try:
			briefing = []
			numberOfSnapshots = 0
			self.__DBW.openCon('/HdHistory')
			try:
				if timestamp:
					operation = "="
				else:
					timestamp = self.lastTimestamp()
					operation = "<="
					
				slashpath = path+"/%%"									
				select =  "select t.timestamp, "\
						+ "t.finalpath, t.operation, sum(t.bsize), "\
						+ "sum(case f.type when 'D' "\
						+ "then 0 else case t.operation when 'DELETED' "\
						+ "then 0 else f.size end end) "\
						+ "from timeline t inner join "\
						+ "file_attribs f on t.id = f.id "\
						+ "where t.timestamp " + operation + " ? "\
						+ "and (t.finalpath = ? or t.finalpath like ?) "\
						+ "group by t.timestamp order by t.timestamp DESC;"
				argv = (timestamp, path, slashpath)
				result = self.__DBW.resultQuery(select, argv)
				if result != [()] and result:
					numberOfSnapshots = len(result) 
					for version in result:
						dataSize = self.getAllDataSizeInVersion(path, version[0])
						briefing.append((	version[0],
											version[1],
											version[2],
											version[3],
											version[4],
											dataSize
										))
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
			return briefing, numberOfSnapshots
		
	##@brief Gets amount of data backed up in a version
	#
	# The briefing some quick info about the versions of a path
	#
	#@param path path of the version
	#@param timestamp time of the versio
	#
	#@return amount (in bytes) of data backed up in a version
	def getAllDataSizeInVersion(self, path, timestamp = None):
		"""
		Gets amount of data backed up in a version
		"""
		if os.path.normpath(path) == "/":
			path = ""
		
		try:
			amountOfData = -1
			self.__DBW.openCon('/HdHistory')
			try:
				if not timestamp:
					timestamp = self.lastTimestamp()

				slashpath = path+"/%%"
				
				select =  "select count(id) "\
						+ "from timeline "\
						+ "where timestamp <= ? "\
						+ "and (finalpath = ? or finalpath like ?)"
				argv = (timestamp, path, slashpath)
				result = self.__DBW.resultQuery(select, argv)
				
				LIMIT = 1000 # Limit to determine real size
				if result != [()] and result and result[0][0] < LIMIT:
					select =  "select sum(case type when 'D' "\
							+ "then 0 else case operation when 'DELETED' "\
							+ "then 0 else size end end) "\
							+ "from (timeline inner join file_attribs "\
							+ "on timeline.id = file_attribs.id) inner join "\
							+ "(select max(timestamp) as time, finalpath as path "\
							+ "from timeline where timestamp <= ? "\
							+ "and (finalpath = ? or finalpath like ?) "\
							+ "group by finalpath order by finalpath) "\
							+ "on (timestamp = time and finalpath = path )"
					argv = (timestamp, path, slashpath)
					result = self.__DBW.resultQuery(select, argv)
				
					if result != [()] and result:
						amountOfData = result[0][0]
						
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
			return amountOfData
			
	##@brief Gets GMT and DST for a given timestamp
	#
	#@param timestamp reference timestamp
	#@return a tuple with GMT and DST for a nearest timestamp to the given
	# timestamp. If GMT and DST are both -1, timestamp has no attributes.
	def getTimeAttributes(self, timestamp):
		"""
		Gets DST and GMT for a given timestamp
		"""
		try:
			result = None
			self.__DBW.openCon('/HdHistory')
			try:
				select =  "select GMT, DST from time_attribs "\
						+ "where timestamp <= ? order by timestamp DESC limit 1"
				argv = (timestamp, )
				result = self.__DBW.resultQuery(select, argv)
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s. Error: %s.',
					select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
			if result and result != [()]:
				return result[0][0], result[0][1]
			else:
				return -1, -1
