# -*- coding: UTF-8 -*-

#Copyright (C) 2008 David Molina

#This file is part of HDLorean.
#
#HDLorean is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#HDLorean is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with HDLorean.  If not, see <http://www.gnu.org/licenses/>.


from DataBaseWrapper import *
from hdlogger import HDLogger
import time


##@brief
#
#Journal database interfaze class
#
#This class implements all the functions needed 
#to control the journal database of HdLorean
class Journal:
	
	__DBW = None #<Contains the DataBaseWrapper.
	__lock = False #<Indicates if the table is locked.
	__logger = None #<Contains the logger.
	__log =None #<Lets write logs messages.
	
	
	##@brief
	#
	#The constructor of the class.	
	def __init__(self):
		self.__DBW =  DataBaseWrapper()
		self.__logger = HDLogger.HDLogger()		
		self.__log = self.__logger.setupHDLogger("hdlorean.db.Journal")	
		
	   
	##@brief 
	#
	#Create the table needed to the journal database.
	def createTable(self):
		try:
			while (self.__lock):				
				time.sleep(1)
			self.__DBW.openCon("/HdJournal")
			try:
				table = 'CREATE TABLE IF NOT EXISTS journal (' 
				table = table + 'id	 INTEGER PRIMARY KEY AUTOINCREMENT, '
				table = table +	'timestamp	 TIMESTAMP   NOT NULL, '
				table = table +	'chtype		 VARCHAR(255)	 NOT NULL, '
				table = table +	'initpath	VARCHAR(255)	 NOT NULL, '
				table = table +	'finalpath	VARCHAR(255)	 NOT NULL, '
				table = table +	'requestor	VARCHAR(255)	 NOT NULL, '
				table = table +	"wrote		CHAR(1)		NOT NULL	DEFAULT '0');"
				index = 'CREATE INDEX IF NOT EXISTS indice ON journal(timestamp) '
				self.__DBW.createTable(table,index)
				result = self.__DBW.resultQuery('select count(*) from journal;',())
				row = result[0]
				if (row[0]==0):
					query = "insert into journal (timestamp,chtype,initpath,finalpath,requestor,wrote)"	
					query = query + " values (0,'SNAPSHOT','/','/','/',1);"
					self.__DBW.noResultQuery(query,())			
				self.__log.info('Journal database created without problems.')   
			except IntegrityError, e:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, e:
				self.__log.critical('SQL error has happened in the database: %s, %s. Error: %s.', table, index, e)
			except ConnError, e:
				self.__log.critical('The connection to the database doesnt exits.')
		finally: 
			self.__DBW.closeCon()
	
		
	##@brief 
	#
	#Inserts in the journal table a new row with the values of the parameters.
	#
	#@param timestamp String with a date in unix format. 
	#@param chtype String with one of the next words: MOVE, CREATE, DELETE, MODIFY.
	#@param init String with the first path of the file.
	#@param final String with the actual path of the file.
	#@param req String with who ordered the change.
	#@param wrote String with an '0'(waiting) or '1'(Backup done)
	def insert(self,timestamp,chtype,init,final,req,wrote):
		try:
			while (self.__lock):				
				time.sleep(1)
			self.__DBW.openCon("/HdJournal")
			try:
				query = "insert into journal (timestamp,chtype,initpath,finalpath,requestor,wrote) values (?,?,?,?,?,?)"
				#query = "insert into journal (timestamp,chtype,initpath,finalpath,requestor,wrote) values ('"+timestamp+"','"+chtype+"','"+init+"','"+final+"','"+req+"','"+wrote+"')"
				self.__DBW.noResultQuery(query,(timestamp,chtype,init,final,req,wrote))
				#query.replace("'","''")
				#self.__DBW.noResultQuery(query)
			except IntegrityError, e:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, e:
				self.__log.critical('SQL error has happened in the database: %s. Error: %s.', query, e)
			except ConnError, e:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
	
	
	##@brief
	#
	#Make an update query to change wrote form 0 to 1.
	#
	#@param idslist List of integers that are the id of the row to update.
	def setWrote(self,idlist):  
		try:
			while (self.__lock):				
				time.sleep(1)
			self.__DBW.openCon("/HdJournal")
			try:
				for id in idlist:
					update = 'update journal set wrote=1  where id=?;'  #< keeps the sql query
					argv = (str(id),)
					self.__DBW.noResultQuery(update,argv)			
			except IntegrityError, e:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, e:
				self.__log.critical('SQL error has happened in the database: %s. Error: %s.', update, e)
			except ConnError, e:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()	
		
		
	##@brief 
	#
	#Insert the special row in the table.
	#
	#@param timestamp String with the date in unix format.
	def insertSpecialRow(self,timestamp):
		try:
			while (self.__lock):				
				time.sleep(1)
			self.__DBW.openCon("/HdJournal")
			try:
				query = "insert into journal (timestamp,chtype,initpath,finalpath,requestor,wrote)"	
				query = query + " values (?,'SNAPSHOT','NULL','NULL','NULL','0');" 
				argv = (timestamp,)
				self.__DBW.noResultQuery(query,argv)
			except IntegrityError, e:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, e:
				self.__log.critical('SQL error has happened in the database: %s. Error: %s.', query, e)
			except ConnError, e:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()

		
	##@brief
	#
	#Search for the rows after the last special row.	
	def getOldest(self):
		try:
			while (self.__lock):				
				time.sleep(1)
			self.__DBW.openCon("/HdJournal")
			try:
				select = "select min(id) from journal where chtype='SNAPSHOT' and wrote = '0'"
				result = self.__DBW.resultQuery(select,())
				if (result != None):
					endId = str(result[0][0])
					select = "select max(id), timestamp, chtype, initpath, finalpath, requestor, wrote from journal where (id > (select max(id) from journal where chtype='SNAPSHOT' and wrote = '1') and id < " + endId + " and chtype<>'SNAPSHOT' and wrote = '0') GROUP BY finalpath ORDER BY id ASC"
				else:
					select = "select max(id), timestamp, chtype, initpath, finalpath, requestor, wrote from journal where (id > (select max(id) from journal where chtype='SNAPSHOT' and wrote = '1') and chtype<>'SNAPSHOT' and wrote = '0') GROUP BY finalpath ORDER BY id ASC"	
				result = self.__DBW.resultQuery(select,())	
				return result
			except SQLError, e:
				self.__log.critical('SQL error has happened in the database: %s. Error: %s.', select, e)
			except ConnError, e:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
	
	
	##@brief
	#
	#Search for the oldest special row with wrote=0 and set it to 1.
	def setWroteSR(self):
		try:
			while (self.__lock):				
				time.sleep(1)
			self.__DBW.openCon("/HdJournal")
			try:
				update = "update journal set wrote=1 where id=(select min(id) from journal where chtype='SNAPSHOT' and wrote=0);"
				self.__DBW.noResultQuery(update,())
			except IntegrityError, e:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, e:
				self.__log.critical('SQL error has happened in the database: %s. Error: %s.', update, e)
			except ConnError, e:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
		
	##@brief
	#
	#Delete all rows in the journal between the first special row and
	#the last one with wrote = 1.
	def freeJournal(self):
		try:
			while (self.__lock):				
				time.sleep(1)				
			self.__DBW.openCon("/HdJournal")
			try:
				delete = "delete from journal where id>1 and id<=(select max(id) from journal where chtype='SNAPSHOT' and wrote='1');"
				self.__DBW.noResultQuery(delete,())
			except IntegrityError, e:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, e:
				self.__log.critical('SQL error has happened in the database: %s. Error: %s.', delete, e)
			except ConnError, e:
				self.__log.critical('The connection to the database doesnt exits.')				
		finally:
			self.__DBW.closeCon()
	
			
	##@brief
	#
	#Lock the journal table.
	def lock(self):
		self.__lock = True
	
	
	##@brief
	#
	#Unlock the journal table.
	def unlock(self):
		self.__lock = False
	
	
	##@brief
	#
	#Search in the journal table for a row with the path specified.
	#
	#@param path String with the path.
	#
	#@return True if there is a row with the path.
	def isPlanified(self,path):
		try:
			while (self.__lock):				
				time.sleep(1)
			self.__DBW.openCon("/HdJournal")
			try:
				select = "select count(*) from journal where initpath=? and wrote = '0'"
				result = self.__DBW.resultQuery(select,(path,))
				row = result[0] 
				if row[0]>0:
					return True
				else:
					return False								
			except SQLError, e:
				self.__log.critical('SQL error has happened in the database: %s. Error: %s.', select, e)
			except ConnError, e:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
