# -*- coding: utf-8 -*-

"""
Copyright (C) 2008 David Molina, Salvador de la Puente, Federico Mon Trotti

This file is part of HDLorean.

HDLorean is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HDLorean is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HDLorean.  If not, see <http://www.gnu.org/licenses/>.
"""

from DataBaseWrapper import DataBaseWrapper, ConnError, SQLError, IntegrityError
from hdlogger import HDLogger

##@brief XDelta3Wrapper support database class.
#
#This class implements all the functions needed 
#to control the support database of XDelta3WrapperDB.
class XDelta3WrapperDB:
	"""
	Lets write logs messages.
	"""
	
	##@brief Contains the DataBaseWrapper.
	#
	__DBW = None
	
	##@brief Contains the logger.
	#
	__logger = None
	
	##@brief Lets write logs messages.
	#
	__log = None
	
	##@brief List with all valid fields of database
	#
	__keys = None
	
	##@brief Default constructor
	#
	def __init__(self):
		self.__DBW =  DataBaseWrapper()
		self.__logger = HDLogger.HDLogger()		
		self.__keys = None
		self.__log = self.__logger.setupHDLogger("hdlorean.db.XDelta3Wrapper")

	##@brief Create the table needed by the XDelta3Wrapper database.
	#
	def createTable(self):
		"""
		Create the table needed by the XDelta3Wrapper database.
		"""
		try:
			self.__DBW.openCon('/HdXDelta3WrapperDB')
			try:
				table = 'CREATE TABLE IF NOT EXISTS history (' 
				table = table + 'id INTEGER PRIMARY KEY AUTOINCREMENT, '
				table = table + 'timestamp TIMESTAMP NOT NULL, '
				table = table + 'operation VARCHAR(255) NOT NULL, '
				table = table + 'btype VARCHAR(255) NOT NULL, ' 
				table = table + 'status VARCHAR(255) NOT NULL, '
				table = table + 'location VARCHAR(255) NOT NULL, '
				table = table + 'path VARCHAR(255) NOT NULL,'
				table = table +	'parent	INTEGER NOT NULL); '
				index = 'CREATE INDEX IF NOT EXISTS indice ON history(timestamp) '
				self.__DBW.createTable(table, index)
				self.__log.info('XDelta3WrapperDB database created without problems.')   
				self.__keys = set([	'id', 'timestamp', 'operation', 'btype',
									'status', 'location', 'path', 'parent'])
			except SQLError, exc:
				self.__log.critical(
					'SQL error has happened in the database: %s, %s. Error: %s.',
					table, index, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
		
	##@brief Inserts in the XDelta3WrapperDB table a new row with the values
	# of the parameters.
	#
	#@param timestamp a unix timestamp.
	#@param operation one of the follow values: MODIFY, MOVE, DELETE, CREATE, CRON.
	#@param btype type of backup: 'INT' or 'DIFF'.
	#@param status status of the backup
	#@param location location of the backup (time label)
	#@param path actual path of the file backed up.
	#@param parent id of the integral of the backup
	def insert(self, 	timestamp,
						operation,
						btype,
						status,
						location,
						path,
						parent = 0):
		"""
		Inserts in the XDelta3WrapperDB table a new row with the values
		of the parameters.
		"""
		try:
			self.__DBW.openCon('/HdXDelta3WrapperDB')
			try:
				insert = "insert into history "\
						 + "(timestamp, operation, btype, status, location, path, parent) "\
						 + "values (?,?,?,?,?,?,?)"
				argv = (timestamp, operation, btype, status, location, path, parent)
				self.__DBW.noResultQuery(insert, argv)
			except IntegrityError:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, exc:
				self.__log.critical(
						'SQL error has happened in the database: %s. Error: %s.',
						insert, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
	
	##@brief Update the timestamp of the path with the old_timestamp
	#
	#@param path string with the path of the file.
	#@param old_timestamp the timestamp in the database.
	#@param new_timestamp the new timestamp.
	def updateTimestamp(self, path, old_timestamp, new_timestamp):
		"""
		Update the timestamp of the path with the old_timestamp
		"""
		try:
			self.__DBW.openCon('/HdXDelta3WrapperDB')
			try:
				update = 'update history set timestamp=? where path=? and timestamp=?;'
				argv = (new_timestamp, path, old_timestamp)
				self.__DBW.noResultQuery(update, argv)				
			except IntegrityError:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, exc:
				self.__log.critical(
						'SQL error has happened in the database: %s. Error: %s.',
						update, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()

	##@brief Deletes a version of a path
	#
	#@param path path of the version to delete (if its a directory, then delete
	# all contents too).
	#@param initialTimestamp Before this timestamp no snapshot will be affected
	#@param endTimestamp After this timestamp no snapshot will be affected
	#
	#@return the next timestamp to move files in the listToMove that
	# it is return too. Fields are:
	# path, timestamp and backup type
	# It return a list of tuples of files to delete. Fields are:
	# path, timestamp, backup type
	def deleteBackups(self, path, initTimestamp = 0, finalTimestamp = 0):
		"""
		Deletes a version of a path
		"""
		# Setup default mode
		if not finalTimestamp:
			finalTimestamp = initTimestamp

		try:
			nextTimestamp = 0
			listToMove = []
			listToDelete = []
			self.__DBW.openCon('/HdXDelta3WrapperDB')
			try:
				slashpath = path + "/%%"
				# Uncheck forbidden backups
				query =   "update history set status='NONE' where "\
						+ "timestamp <= ? and timestamp >= ? and "\
						+ "(path=? or path like ?) and btype='INT';"
				argv = (str(finalTimestamp), str(initTimestamp), path, slashpath)
				self.__DBW.noResultQuery(query, argv)
		
				# Look for inmediatly next timestamp to de finalTimestamp
				query =   "select timestamp from history where "\
						+ "timestamp > ? and (path=? or path like ?) "\
						+ "order by timestamp ASC limit 1;"
				argv = (str(finalTimestamp), path, slashpath)
				nextTimestamp = self.__DBW.resultQuery(query, argv)

				# If it exists (we are not deleting last backup)
				if nextTimestamp:
					nextTimestamp = nextTimestamp[0][0]

					# Select versions of path with nextTimestamp
					subquery = "select path from history where timestamp = ?"
					# Look for last versions in the timestamp that there are
					# no versions of them in nextTimestamp.
					query = "select id,path,max(timestamp),btype from history "\
						   + "where timestamp <= ? and timestamp >= ? and "\
						   + "(path=? or path like ?) and path not in "\
						   + "(" + subquery + ") group by path;"
						   
					argv = (finalTimestamp,
							initTimestamp,
							slashpath,
							nextTimestamp)
					toPromote = self.__DBW.resultQuery(query, argv)
					
					# If there are files to promote in time
					if toPromote:
						listToMove = [x[1:] for x in toPromote]
						for entry in toPromote:
							query = "update history set timestamp=? where id = ?;"
							argv = (nextTimestamp, entry[0])
							self.__DBW.noResultQuery(query, argv)
			
					# Look for parents of versions at nextTimestamp (including
					# recently promoted backups)
					subquery = "select parent from history where timestamp = ? "\
							  + "and (path=? or path like ?)"
					query = "update history set status = 'forbidden' where id in "\
							 + "(" + subquery + ");"
					argv = (nextTimestamp, path, slashpath)
					self.__DBW.noResultQuery(query, argv)
					
				# Delete (and add to deleteList) all backups not forbidden
				query = "select path,timestamp,btype from history where "\
						 + "status != 'forbidden' and timestamp <= ? and "\
						 + "timestamp >= ? and (path=? or path like ?);"
				argv = (finalTimestamp, initTimestamp, path, slashpath)
				listToDelete = self.__DBW.resultQuery(query, argv)
				
				query = "delete from history where status != 'forbidden' and "\
						 + "timestamp <= ? and timestamp >= ? and "\
						 + "(path=? or path like ?);"
				argv = (finalTimestamp, initTimestamp, path, slashpath)
				self.__DBW.noResultQuery(query, argv)
			except IntegrityError:
				self.__log.info('Integrity error has happened in the database.')
			except SQLError, exc:
				self.__log.info(
						'SQL error has happened in the database: %s. Error: %s.',
						query, exc)
			except ConnError:
				self.__log.info('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			return nextTimestamp, listToMove, listToDelete
	
	##@brief Search for the latest backup of a file.
	#
	#@param timestamp date in unix timestamp format.
	#@param files list of strings, each string is the complete path of the file.
	# Example: ['/home/Documents/example.py','/home/example2.txt']
	#@param integral if true, look only for integral backups.
	#@param recursive optional boolean (by default is False). Set it true
	# if you want to do a recursive search (it means, recover the content of
	# a path)
	#
	#@return A list with the rows selects in the same order of paths in the list
	# files, if a path has not result in that position will be 
	# an empty sequence like ()
	def lastBackUp(self, timestamp, files, integral, recursive = False):
		"""
		Search for the latest backup of a file.
		"""
		lastBackups = []
		try:
			self.__DBW.openCon('/HdXDelta3WrapperDB')
				
			try:
				btype = ""
				if integral:
					btype = "and btype='INT'"
				
				for path in files:
					if recursive: 
						slashpath = path + "/%%"
						select =  "select * from history where "\
								+ "(path = ? or path like ? or path like ?) "\
								+ btype + " and timestamp <= ? "\
								+ "order by timestamp DESC;"
						argv = (path, path, slashpath, timestamp)
					else:
						select =   "select * from history where path =? "\
								 + btype + " and timestamp<=? order by timestamp DESC;"
						argv = (path, timestamp)
						
					result = self.__DBW.resultQuery(select, argv)
					if (result != None):
						lastBackups.append(result[0])
			except SQLError, exc:
				self.__log.critical(
						'SQL error has happened in the database: %s. Error: %s.',
						select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			return lastBackups
	
	##@brief Return an entry by its ID
	#
	#@param identifier id of the row
	#
	#@return a DB entry with ID given
	def getByID(self, identifier):
		"""
		Return an entry by its ID
		"""
		result = []
		try:
			self.__DBW.openCon('/HdXDelta3WrapperDB')	
			try:
				select = "select * from history where id = ?;"
				argv = (identifier, )
				result = self.__DBW.resultQuery(select, argv)
			except SQLError, exc:
				self.__log.critical(
						'SQL error has happened in the database: %s. Error: %s.',
						select,	exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
		finally:
			self.__DBW.closeCon()
			
			if result and result != [()]:
				return result[0]
			else:
				return ()
						
	##@brief Search for the backups before timestamp and like path.  
	#
	#@param timestamp unix timestamp to search before.
	#@param path path of the version.
	#@param recursive set to True if you want to select all files under the path.
	#@param latest set to True in order to select the backup nearest (but always
	# before) timestamp parameter. Set it to False if you want all backups 
	# with lower timestamp. 
	#@param btype type of backup to search. By default, ist value is 'all'.
	#@param grouped indicate a field to group by in the select.
	#
	#@return array with backups before timestamp and like path.
	def getSnapshots(self, 	timestamp,
							path,
							recursive,
							latest,
							btype = 'all',
							grouped = ''):
		"""
		Search for the backups before timestamp and like path.  
		"""
		result = None
		listArgs = []
		try:
			self.__DBW.openCon('/HdXDelta3WrapperDB')
			try:
				if (not latest):
					timecondition = '='
				else:
					timecondition = '<='

				# Get all contents too
				if (recursive):
					slashpath = path + "/%%"
					comparison = "select * from history where timestamp "\
								+ timecondition + "? and (path = ? or path like ?)"
					listArgs.extend([timestamp, path, slashpath])
				else: 
					comparison = "select * from history where timestamp "\
								+ timecondition + "? and path like ?"
					listArgs.extend([timestamp, path])
				
				# Type of backups to look for
				if (btype != 'all'):
					comparison = comparison + " btype = ?"
					listArgs.append(btype)

				# Group by
				if grouped != '' and grouped in self.__keys:
					comparison = comparison + " group by " + grouped
					
				select = comparison + " order by timestamp DESC;"
				result = self.__DBW.resultQuery(select, tuple(listArgs))
			except IntegrityError:
				self.__log.critical('Integrity error has happened in the database.')
			except SQLError, exc:
				self.__log.critical(
						'SQL error has happened in the database: %s. Error: %s.',
						select, exc)
			except ConnError:
				self.__log.critical('The connection to the database doesnt exits.')
				
		finally:
			self.__DBW.closeCon()
			
			if result and result != [()]:
				return result
			else:
				return []