# -*- coding: UTF-8 -*-

#Copyright (C) 2008 Salvador de la Puente

#This file is part of HDLorean.
#
#HDLorean is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#HDLorean is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with HDLorean.  If not, see <http://www.gnu.org/licenses/>.


import logging
import logging.handlers
import os

##@brief HDLogger is a Borg (monostate DP) class that configures properly any hdlorean logger.
# The setupHDLogger method lets developer to get a Python logger properly setted to write into
# HD-Lorean log files. There are four log files inside "~/.hdlorean/logs":
#
# log		- where only last logs are written
# debug.log - where all logs are written
# error.log - where only ERROR or greater logs are written
# info.log  - where all logs execept DEBUG logs are written
# 
# The last three los are rotative, it means that it files reach a maximum size, logger open a
# new file to write into. Until three old log files are stored in this way "*.log.<number>"
# where number can be 1, 2 or 3.
#
class HDLogger:
	
	##@brief Shared state dictionary for Borg monostate DP
	#
	__shared_state = {}
	
	##@brief Indicates if root logger setup has been performed or not
	#
	__setupPerformed = False
	
	##@brief Stores which of loggers has been setted
	#
	__loggerSetupPerformed = []
	
	##@brief Format string
	#
	__logFormatString = "%(asctime)s (%(module)s:%(filename)s #%(lineno)d) %(levelname)s > %(message)s"
	
	##@brief Formatter object for logs
	#
	__logFormat = None
	
	##@brief Constructor of the Borg monostate DP
	#
	def __init__(self):
		self.__dict__ = self.__shared_state
		self.__logFormat = logging.Formatter(self.__logFormatString)
	
	##@brief Setup HD-Lorean logger file and 
	#
	# @param strLoggerName name of logger
	def setupHDLogger(self,strLoggerName):
		if not self.__setupPerformed:
			# Setup path # FIXME Recuperar desde el config manager
			homePath = os.getenv("HOME", "~")
			logDirPath = os.path.normpath(os.path.join(homePath,".hdlorean","logs"))
			
			# Log files
			self.__debugLogPath = os.path.normpath(os.path.join(homePath,".hdlorean","logs","debug.log"))
			self.__infoLogPath = os.path.normpath(os.path.join(homePath,".hdlorean","logs","info.log"))
			self.__errorLogPath = os.path.normpath(os.path.join(homePath,".hdlorean","logs","error.log"))
			self.__lastLogPath = os.path.normpath(os.path.join(homePath,".hdlorean","logs","log"))
			
			# Make dirs
			if not os.path.exists(logDirPath):
				os.makedirs(logDirPath)
				
			# Setup root logger
			logging.basicConfig(filename=self.__lastLogPath,filemode='w',format=self.__logFormatString,level=10)	
			
			self.__setupPerformed = True;
			
		# Setup custom logger
		logger = logging.getLogger(strLoggerName)
		if not (strLoggerName in self.__loggerSetupPerformed):
			
			# Configure handlers
			MEGA = 1048576
			
			debugLogHndl = logging.handlers.RotatingFileHandler(self.__debugLogPath,'a',MEGA,3)
			debugLogHndl.setFormatter(self.__logFormat)
			debugLogHndl.setLevel(10)
			
			infoLogHndl = logging.handlers.RotatingFileHandler(self.__infoLogPath,'a',MEGA,3)
			infoLogHndl.setFormatter(self.__logFormat)
			infoLogHndl.setLevel(20)
			
			errorLogHndl = logging.handlers.RotatingFileHandler(self.__errorLogPath,'a',MEGA,3)
			errorLogHndl.setFormatter(self.__logFormat)
			errorLogHndl.setLevel(30)
			
			# Add handlers
			logger.addHandler(debugLogHndl)
			logger.addHandler(infoLogHndl)
			logger.addHandler(errorLogHndl)
			
			# Add logger to setup performed set
			self.__loggerSetupPerformed.append(strLoggerName)
		
		return logger
		
