# -*- coding: UTF-8 -*-

"""
Copyright (C) 2008 Salvador de la Puente, Federico G. Mon Trotti

This file is part of HDLorean.

HDLorean is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HDLorean is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HDLorean.  If not, see <http://www.gnu.org/licenses/>.
"""

import time
import mimetypes
import os
import subprocess

from db import Journal
from db import History
from watcher import ConfigFileManager
from snapshotManager import IStorageWrapper
from snapshotManager import XDelta3Wrapper

from stat import ST_SIZE, ST_MODE, ST_ATIME, ST_CTIME, ST_MTIME

##@brief SnapshotManager class encapsulate all funtionallity related to
# versions management including creation, deletion and restoration.
#
# It implements a monostate (Borg) design pattern (DP). It consists in a shared
# data structure shared by all instances of this class.
#
# @author Salvador de la Puente González, Federico G. Mon Trotti
# @contact neo.salvador@gmail.com
class SnapshotManager:
	"""
	SnapshotManager class encapsulate all funtionallity related to versions 
	management including creation, deletion and restoration.
	"""
	
	##@brief Shared state dictionary for Borg monostate DP. (See class commentary)
	#
	__sharedState = {}
	
	##@brief To control initialization.
	#
	__done = False
	
	##@brief Object that comunicates with storage device.
	#
	__storageManager = None

	##@brief Object to access history db.
	#
	history = None

	##@brief Object to access journal db.
	#
	__journal = None

	##@brief Object to access configfile manager.
	#
	__configFile = None

	##@brief Backup path where versions are stored.
	#
	__backupsPath = None

	##@brief Wrapper that is making backups.
	#
	__actualWrapper = None

	##@brief Dictionary with all wrappers associated by their names.
	#
	__storageDictionary = {}

	##@brief Default constructor.
	#
	def __init__(self):
		self.__dict__ = self.__sharedState
		
		if not self.__done:
			# To access journal and history
			self.__journal = Journal.Journal()
			self.history = History.History()
			self.history.createTables()
	
			# To get and set HD Lorean settings
			self.__configFile = ConfigFileManager.ConfigFileManager()
			self.__backupsPath = self.__configFile.getBackupDirectory()
		
			self.__done = True		

	##@brief Reload configfile manager settings.
	#
	def reloadConfiguration(self):
		"""
		Reload configfile manager settings
		"""
		self.__backupsPath = self.__configFile.getBackupDirectory()

	##@brief Gets actual backup path.
	#
	#@return actual backup path.
	def getBackupPath(self):
		"""
		Gets actual backup path
		"""
		return self.__backupsPath

	##@brief Setups the proper storage manager.
	#
	# @param manager indicates which storage manager must be used.
	# By default, manager is setted to 'xdelta3'.
	def setup(self, manager = "xdelta3"):
		"""
		Setups the proper storage manager.
		"""
		self.__actualWrapper = manager
		if manager == "rdiff-backup":
			self.__storageManager = RdiffBackupWrapper.RdiffBackupWrapper(self)
		elif manager == "xdelta3":
			self.__storageManager = XDelta3Wrapper.XDelta3Wrapper(self)

		self.__storageDictionary[manager] = self.__storageManager

	##@brief Gets a storage manager by its name.
	#
	#@param managerName identify which storageManager is required.
	# It can be 'xdelta3' or 'dummy'.
	def __getStorageManager(self, managerName):
		"""
		Gets a storage manager by its name.
		"""
		if managerName not in self.__storageDictionary:
			sManager = None
			if managerName == "rdiff-backup":
				sManager = RdiffBackupWrapper.RdiffBackupWrapper(self)
			elif managerName == "xdelta3":
				sManager = XDelta3Wrapper.XDelta3Wrapper(self)

			self.__storageDictionary[managerName] = sManager

		return self.__storageDictionary[managerName]
	

	##@brief Gets present timestamp, GMT difference and DST indicator.
	#
	#@return a tuple with timestamp, GMT and DST.
	def __getTimestamp(self):
		"""
		Gets present timestamp, GMT difference and DST indicator.
		"""
		now = time.localtime()
		timestamp = int(time.time())
		return timestamp, time.timezone, now[8]
	
	##@brief Obtain MIME Type for a file.
	#
	#@param  filepath path of the file to be analyzed. 
	#@return MIME Type of a file or empty string if it cannot be determined.
	def __getMimeType(self, filepath):
		"""
		Obtain MIME Type for a file.
		"""
		cmd = "file -i '%s'" % filepath
		try:
			child = subprocess.Popen(	[cmd],
										stdin=subprocess.PIPE,
										stdout=subprocess.PIPE,
										close_fds=True,
										shell=True
									)
			stdout = child.stdout.read()
			mime = stdout.split(': ')[-1][:-1]
		except OSError:
			mimetypes.init()
			mime = mimetypes.guess_type(filepath)
			
		if not mime:
			return ""
		else:
			return mime
			
	   
	##@brief Makes a backup for a set of files.
	#
	#@param fileList a list of tuples with 
	# id, timestamp, operation, init_path, final_path, requestor, wrote
	# sended from the Scheduler.
	#@param integral if it's setted true, force to make integral backups of
	# everything. By default it is false.
	def makeBackups(self, fileList, integral = False):
		"""
		Makes a snapshot of a set o files.
		"""
		# Generates timestamp for the snapshot
		# REMEMBER: self.__getTimestamp() returns a tuple with three elements
		timeAttribs = self.__getTimestamp()
		# Insert time attributes
		self.history.insertTimeAttributes(timeAttribs)
		
		# Log info
#		self.__log.info("==== STARTING SNAPSHOT PROCESS ('%s') ====" % snapPath)
			
		# For each element
		for actualFile in fileList:
			# Make individual backup
			bSize = self.__storageManager.makeBackup(	timeAttribs,
														actualFile,
														integral
													   )
			
			# Add new version to history			
			#FIXME: Extract UUID and pass where the zero is passed
			versionInfo =	timeAttribs[0], actualFile[3], actualFile[4], 0, \
							actualFile[2], actualFile[5], bSize, \
							self.__actualWrapper
			fileID = self.history.insertNewVersion(versionInfo)
			
			# Extract file attributes and insert into database
			try:
				stats = os.stat(actualFile[4])
			except OSError:
				stats = None
			
			fileXPerms = 0 #FIXME: Replace me by real extended perms
			fileMimeType = self.__getMimeType(actualFile[4])
			fileType = (os.path.isdir(actualFile[4]) and ['D'] or ['F'])[0]
			
			# TODO: Si se va a borrar, sacar resto de información de la última
			# entrada de la base de datos.
			fileAttribs = fileID, 0, 0, fileXPerms, fileMimeType, 0, 0, 0, fileType
			if stats:
				fileAttribs = 	fileID, stats[ST_SIZE], stats[ST_MODE], \
								fileXPerms,	fileMimeType, stats[ST_ATIME], \
								stats[ST_MTIME], stats[ST_CTIME], fileType
			
			self.history.insertFileAttributes(fileAttribs)
			
			# Set journal operation as wrote			
			self.__journal.setWrote([actualFile[0]])								
		# Log info
#		self.__log.info("==== END OF SNAPSHOT PROCESS ====")
		return

	
	##@brief Recover a file or a directory to a specific path at time given.
	#
	#@param timestamp timestamp at the file or directory must be recovered.
	#@param originalPathList list of paths for the file(s) or directories
	# to recover.
	#@param recoverPath path where recovered file(s) are placed.
	#@return actual recover path where files(s) has been recovered.
	def recoverBackups(self, timestamp, originalPathList, recoverPath):
		"""
		Recover a file or a directory to a specific path.
		"""
		# Get wrapper for that timestamp
		wrapperName = self.history.whichWrapper(timestamp)
		if wrapperName:
			wrapper = self.__getStorageManager(wrapperName)
			return wrapper.recoverBackups(	timestamp,
											originalPathList,
											recoverPath
										  )
		#TODO: Raise an exception
			
	##@brief Delete an interval of versions of a path between two timestamps.
	#
	#@param path path of the versions to delete
	#@param initTimestamp low timestamp of the interval. By default, it's 0.
	#@param finalTimestamp hight timestamp of the interval. By default, it's 0.
	def deleteBackups(self, path, initTimestamp = 0, finalTimestamp = 0):
		"""
		Delete an interval of versions of a path between two timestamps.
		"""			
		if not initTimestamp and not finalTimestamp:
			initTimestamp = 0
			finalTimestamp = self.lastTimestamp()
			
		elif not finalTimestamp:
			finalTimestamp = initTimestamp		
		
		result = self.history.getWrappersBetween(	initTimestamp,
													finalTimestamp,
													path
												  )
		self.history.deleteVersions(path, initTimestamp, finalTimestamp)
		if result:
			for wrapperName in result:	
				wrapper = self.__getStorageManager(wrapperName[0])
				wrapper.deleteBackups(path, initTimestamp, finalTimestamp)
				
	##@brief Gets basic info about all contents versioned once of a directory.
	#
	#@param path path to recover its contents
	#@param oneLevel if it is true then, method just recover sons of
	# one deep level.
	#@return a list with paths, type of each path and MIME Type.
	def getDeepDirList(self, path, oneLevel = False):
		"""
		Gets basic info about all contents versioned once of a directory.
		"""
		return self.history.getDeepDirList(path, oneLevel)
	
	##@brief Gets the historical of versions for a path.
	#
	# The historical of a path is a list with all versions tracked
	# over the path and same info of getVersionInfo.
	#
	#@param path path of file or directory. If path is a directory then,
	# operations over sons of that directory are also shown inside historical.
	#@param timestamp by default is None, but given a unix timestamp, returns
	# only that version info.
	#
	#@return a list of tuples with the next information (it can be only one if
	# a timestamps is specified):
	# ID
	# timestamp
	# init path
	# operation path
	# location
	# operation
	# requestor
	# backup size
	# name of sm
	# file size
	# permissions
	# x permissions
	# mime type
	# last access
	# last modif
	# creation time
	# type (F or D)
	def getVersionsInfo(self, path, timestamp = None):
		"""
		Gets the historical of versions for a path.
		"""
		return self.history.getVersionsInfo(path, timestamp)
	
	##@brief Gets a quick briefing of all versions of a path
	#
	# The briefing consists in some quick info about the versions of a path.
	#
	#@param path path of file or directory. If path is a directory then,
	# operations over sons of that directory are also shown inside briefing.
	#
	#@return a pair with a list of tuples with the next information
	# timestamp
	# operation path
	# operation
	# backup size
	# file size
	# amount of backed up data*
	#
	# And total number of backups too.
	#
	# If version has more than 1000 files, -1 is returned
	# TODO: Extends note with average of data size and composition	
	def getVersionsBriefing(self, path, timestamp = None):
		"""
		Gets a quick briefing of all versions of a path
		"""
		return self.history.getVersionsBriefing(path, timestamp)
	
	##@brief Gets GMT and DST for a given timestamp
	#
	#@param timestamp reference timestamp
	#@return a tuple with GMT and DST for a nearest timestamp to the given
	# timestamp. If GMT and DST are both -1, timestamp has no attributes.
	def getTimeAttributes(self, timestamp):
		"""
		Gets DST and GMT for a given timestamp
		"""
		return self.history.getTimeAttributes(timestamp)
	
	##@brief Gets the latest timestamp in the data base.
	#
	#@return 0 if there is not result or a timestamp.
	def lastTimestamp(self):
		"""
		Gets the latest timestamp in the data base.
		"""
		return self.history.lastTimestamp()
	
	##@brief Gets the oldest timestamp in the data base.
	#
	#@return 0 if there is not result or a timestamp.
	def firstTimestamp(self):
		"""
		Gets the oldest timestamp in the data base.
		"""
		return self.history.firstTimestamp()
	
	##@brief Search for the next timestamp of the param.
	#
	#@param timestamp reference timestamp
	#@return 0 if there isnt result or next timestamp.
	def prevTimestamp(self, timestamp):
		"""
		Search for the next timestamp of the param.
		"""
		return self.history.prevTimestampst(timestamp)
	
	##@brief Search for the previous timestamp of the param.
	#
	#@param timestamp reference timestamp
	#@return None if there isnt result or the previous timestamp.
	def nextTimestamp(self, timestamp):
		"""
		Search for the previous timestamp of the param.
		"""
		return self.history.nextTimestamp(timestamp)

	##@brief Gets timestamps in an interval.
	#
	#@param initTimestamp reference initial timestamp
	#@param finalTimestamp reference final timestamp
	#@return timestamps in an interval.
	def getTimestampsBetween(self, initTimestamp, finalTimestamp):
		"""
		Gets timestamps in an interval.
		"""
		return self.history.getTimestampsBetween(initTimestamp, finalTimestamp)
		
	##@brief Gets amount of data backed up in a version
	#
	# The briefing some quick info about the versions of a path
	#
	#@param path path of the version
	#@param timestamp time of the versio
	#
	#@return amount (in bytes) of data backed up in a version
	def getAllDataSizeInVersion(self, path, timestamp = None):
		"""
		Gets amount of data backed up in a version
		"""
		return self.history.getAllDataSizeInVersion(path, timestamp)
