# -*- coding: UTF-8 -*-

"""
Copyright (C) 2008 Salvador de la Puente, Federico Mon Trotti

This file is part of HDLorean.

HDLorean is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HDLorean is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HDLorean.  If not, see <http://www.gnu.org/licenses/>.
"""

import time
import os
import string
import shutil
import thread
import threading
import subprocess
from stat import ST_SIZE
from db import XDelta3WrapperDB
from snapshotManager import IStorageWrapper
from hdlogger import HDLogger

##@brief Inherits from IStorageWrapper. XDeltaWrapper uses a binary-diff tool
# called xdelta3 to perform more efficiently in size backups of the system.
#
class XDelta3Wrapper(IStorageWrapper.IStorageWrapper):
	"""
	Inherits from IStorageWrapper. XDeltaWrapper uses a binary-diff tool
	called xdelta3 to perform more efficiently in size backups of the system.
	"""
	
	##@brief Constant name of the wrapper
	#
	__NAME = 'xdelta3'

	##@brief Constant of minimun reduction ratio for the diff files.
	#
	__MINIMUM_REDUCTION = 80

	##@brief SnapshotManager instance.
	#
	__SM = None
	
	##@brief Database object. Database maintains vital info about backups and
	# some redundant info in order to free History db traffic. 
	#
	__db = None
		
	##@brief Logging tool
	#
	__logger = None
	
	##@brief Log object
	#
	__log = None

	
	##@brief Lock for recover rutine in order to block creation of the same 
	# directory twice.
	#
	__lock = None

	##@brief Initializes the manager, including self.__db object and util paths
	#
	def __init__(self, smObject):
		# Reference to SnapshotManager		
		self.__SM = smObject

		# Locks
		self.__lock = threading.Lock()

		# Logger
		self.__logger = HDLogger.HDLogger()
		self.__log = self.__logger.setupHDLogger(
									'hdlorean.SnapshotManager.XDelta3Wrapper')
		
		# To access wrapper table
		self.__db = XDelta3WrapperDB.XDelta3WrapperDB()
		self.__db.createTable()

		# Reduction ratio setted to 80%
		self.__MINIMUM_REDUCTION = 80
		
	##@brief Converts a path timeString into a human readable format
	#
	#@param timeString path timestring with format YYYY/MM/DD/HH:MM:SSGMT+X[DST]
	#@return a human readable timestring
	def __humanizeTimeString(self, timeString):
		"""
		Converts a path timeString into a human readable format
		"""
		timeString = timeString.replace("/", "-", 2)
		timeString = timeString.replace("/", " ", 1)
		return timeString

	##@brief Receives an optional timestamp and return a backup path based on it
	#
	#@param unixTimeStamp optional parameter. 
	# If its provided, path is based on that timestamp,
	# else, path is based on the present tiemstamp.
	#
	#@returns a tuple with a string with the absolute snapshot path based on the
	# tiemstamp and a human-readable timestamp.
	#
	# It can return 0 and the string 'never' if timestamp is not in database
	def __getPath(self, unixTimeStamp = None):
		"""
		Receives an optional timestamp and return a backup path based on it
		"""
		# If unixTimestamp is setted
		if unixTimeStamp != None:
			# Get GMT (difference beetwen time zone an Grenwich Time)
			# and DST (day saving time, summer time)
			notGMT, isDST = self.__SM.history.getTimeAttributes(unixTimeStamp)
			
			if notGMT == -1 and isDST == -1:
				return 0, 'never'
			
			notGMT = int(notGMT)
			isDST = int(isDST)

			# Human readable TimeStamps
			humanTimeStamp = time.gmtime(unixTimeStamp - notGMT)
		
		# If unixTimeStamp is not setted, get it now
		else:			
			humanTimeStamp = time.localtime()
			notGMT = time.timezone
			isDST = humanTimeStamp[5]

		humanTimeStamp = list(humanTimeStamp[0:6])
		# Sets time-folder format
		timeFormat = '%02d/%02d/%02d/%02d:%02d:%02d'
		# Calculates GMT difference and add it to timeString		

		if notGMT: 
			GMT = -(notGMT/3600)
			timeFormat += 'GMT%+d' % GMT
		
		# If it is in DST, add it to timestring and add 1 hour
		if isDST:
			humanTimeStamp[3] += 1
			timeFormat += 'DST'

		timeString = timeFormat % tuple(humanTimeStamp)
		return 	(os.path.join(self.__SM.getBackupPath(), timeString), 
				timeString)
	
	##@brief Return a snapshot path based on timestamp, GMT difference and DST
	#
	#@param unixTimeStamp GMT unix timestamp
	#@param notGMT GMT difference in seconds
	#@param isDST DST mode 
	#@returns a tuple with a snapshot path and a human-readable timestamp.
	def __buildPath(self, unixTimeStamp, notGMT, isDST):
		"""
		Return a snapshot path based on timestamp, GMT difference and DST
		"""
		# Get time
		humanTimeStamp = time.gmtime(unixTimeStamp - notGMT)
		humanTimeStamp = list(humanTimeStamp[0:6])
		timeFormat = '%02d/%02d/%02d/%02d:%02d:%02d'
		# Calculates GMT difference and add it to timeString		
		if notGMT: 
			GMT = -(notGMT/3600)
			timeFormat += 'GMT%+d' % GMT
		# If it is in DST, add it to timestring
		if isDST:
			timeFormat += 'DST'
			humanTimeStamp[3] += 1
		timeString = timeFormat % tuple(humanTimeStamp)
		return (os.path.abspath(os.path.join(self.__SM.getBackupPath(), timeString)),
				timeString)
	
	##@brief Parses verbose xdelta3 output to find reduction ratio
	#
	#@param xdeltastdout output of 'xdelta3 -v'-like command
	#@return a string with the reduction ratio (it can be inf to indicate infinite)
	def __parseReductionPercent(self, xdeltaStdout):
		"""
		Parses verbose xdelta3 output to find reduction ratio
		"""
		number = 101
		if xdeltaStdout.find('(') > -1:
			number = xdeltaStdout.split('(')[-1].split('%')[0].split('.')[0]
		return number
	
	
	##@brief Encode a patch with xdelta3
	#
	#return level of compression of the patch
	def __encodeXDelta3(self, source, path, patch):
		"""
		Encode a patch with xdelta3
		"""
		cmd = 'xdelta3 -v encode -s \"%s\" \"%s\" \"%s\"' % (source, path, patch)
		try:
			child = subprocess.Popen(	[cmd],
										stdin = subprocess.PIPE,
										stdout = subprocess.PIPE,
										close_fds = True,
										shell = True,
										stderr = subprocess.PIPE
									)
			
			stdout = child.stderr.read()
			reductionPercent = self.__parseReductionPercent(stdout)
			return reductionPercent
		except OSError:
			# TODO: Raise an exception indicating that xdelta was not found
			self.__log.critical("Xdelta3 could not be called.")
			return "inf"


	##@brief Decode a patch with xdelta3
	#
	def __decodeXDelta3(self, source, path, patch):
		"""
		Decode a patch with xdelta3
		"""
		cmd = 'xdelta3 -v decode -f -s \"%s\" \"%s\" \"%s\"' % (source, path, patch)
		try:
			subprocess.Popen(	[cmd],
								stdin = subprocess.PIPE,
								stdout = subprocess.PIPE,
								close_fds = True,
								shell = True,
								stderr = subprocess.PIPE
							)
		except OSError:
			# TODO: Raise an exception indicating that xdelta was not found
			self.__log.critical("Xdelta3 could not be called.")
			
	##@brief Process a regular file backup.
	#
	# This function is neccesary in order to simplify makeBackups behaviour.
	# So its recomended to see makeBackups() documentation first.
	#
	# The function try to make a difference patch between new file and an old
	# version but, if that old version doesn't exist or if size of new patch is
	# greater than the original filesize, then an identical copy
	# is made instead the patch.
	#
	def __processFile(self, operation,
							timestamp,
							filePath,
							boolIntegral,
							tmpBackupFilePath):
		"""
		Process a regular file backup.
		"""
		lastIntegralID = 0
		backupSize = 0
		backupMode = 'NN'
		# Depends on operation
		if operation != 'DELETE':
			# Gets last integral
			lastIntegral = self.__db.lastBackUp(str(timestamp), [filePath], True)
			
			# If someone is found...
			if not boolIntegral and lastIntegral:						
				# Sets backup mode to DIFF
				backupMode = 'DIFF'
				backupFilePath = tmpBackupFilePath + '.diff'
				
				# Gets last integral path
				lastIntegralID = lastIntegral[0][0]
				lastIntegralPath = os.path.join(self.__SM.getBackupPath(), 
												lastIntegral[0][5])
				lastIntegralPath = os.path.join(lastIntegralPath,
												filePath[1:] + '.int')
				
				# Make differential file using xdelta3 to generate diff file
				# and look for reduction ratio. Reduction ratio let us to
				# determine if differential file is admisible instead of 
				# the integral backup.
				reductionPercent = self.__encodeXDelta3(lastIntegralPath,
														filePath,
														backupFilePath)					
				
				if 	(reductionPercent == 'inf' or
					 int(reductionPercent) > self.__MINIMUM_REDUCTION): 

					boolIntegral = True
					try:
						os.remove(backupFilePath)
					except OSError:
						if not os.path.exists(filePath):
							self.__log.warning(
								"Possible 'DELETE' operation missed over '%s'."
								% filePath)
						else:
							self.__log.error(
								"Copy '%s' to '%s' did not performed!"
								% (filePath, backupFilePath))
					
			# If there isn't last integral, force integral...
			else:
				boolIntegral = True
					
			# If integral backup is needed
			if boolIntegral:
				# Sets backup mode to INT
				backupMode = 'INT'
				backupFilePath = tmpBackupFilePath + '.int'
				try:
					shutil.copy2(filePath, backupFilePath)
				except IOError:
					if not os.path.exists(filePath):
						#TODO: Interrumpir y lanzar excepción
						self.__log.warning(
							"Possible 'DELETE' operation missed over '%s'."
							% filePath)
					else:
						self.__log.error(
							"Copy '%s' to '%s' did not performed!"
							% (filePath, backupFilePath))
			try:
				backupSize = os.stat(backupFilePath)[ST_SIZE]
			except OSError:
				backupSize = -1
				
		return lastIntegralID, backupMode, backupSize
		
	##@brief Process a directory backup.
	#
	# This function is neccesary in order to simplify makeBackups behaviour.
	# So its recomended to see makeBackups() documentation first.
	#
	def __processDirectory(	self,
							operation,
							tmpBackupFilePath):
		"""
		Process a directory backup.
		"""
		# Set backup mode
		backupMode = 'DIR'
		
		# Depends on operation
		if operation == 'CREATE' or operation == 'MOVE':
			if not os.path.exists(tmpBackupFilePath):
				os.makedirs(tmpBackupFilePath)
				
		return backupMode

	##@brief Makes a backup for a file given
	#
	#
	#@param timeAttribs timestamp, GMT and DST of the backup
	#@param actualFile tuple with
	# id, timestamp, operation, init_path, final_path, requestor, wrote
	# sended from the Scheduler module.
	#
	#@param boolIntegral if it's setted true, force to make an integral backup.
	def makeBackup(self, timeAttribs, actualFile, boolIntegral = False):
		"""
		Makes a snapshot of a set o files now.
		"""
		# Build up backup root directory
		rootBackupPath, label = self.__buildPath(
												timeAttribs[0],		# Timestamp
												timeAttribs[1],		# GMT
												timeAttribs[2])		# DST
		try:
			os.makedirs(rootBackupPath)
		except OSError:
			self.__log.info("The snapshot path '%s' already exists!" % rootBackupPath)
						
		# Get file attributes
		operation = actualFile[2]
		filePath = actualFile[4]
				
		# Build up backup concrete directory
		#REMEMBER: Posfix [1:] is required to delete symbol '/' from the absolute path
		parentPath = os.path.split(filePath)[0]
		backupPath = os.path.join(rootBackupPath, parentPath[1:])
		tmpBackupFilePath = os.path.join(rootBackupPath, filePath[1:])
		if not os.path.exists(backupPath):
			try:
				os.makedirs(backupPath)
			except OSError:
				self.__log.error(
							"The backup path '%s' could not been created!"
							% rootBackupPath)

		# Initialize variables
		parent = 0
		backupSize = 0
		backupMode = 'NN' #REMEMBER: NN <--> Not Necessary backup
		
		# It is a file (not a directory)
		if not os.path.isdir(filePath):
			parent, backupMode, backupSize = self.__processFile(
															operation,
															timeAttribs[0],
															filePath,
															boolIntegral,
															tmpBackupFilePath)
			
		# A directory
		else:
			backupMode = self.__processDirectory(operation, tmpBackupFilePath)
		
		# Write this file into database (as wrote)
		fileLength = 0
		if os.path.exists(filePath):
			fileLength = os.stat(filePath)[ST_SIZE]
		elif operation != 'DELETE':
			#TODO: Interrumpir y lanzar excepción
			self.__log.warning("Possible 'DELETE' operation missed.")
			self.__log.error(	"No statistics available: file '%s' does not exist!"
								% filePath)
			
		self.__db.insert(	timeAttribs[0],
							operation,
							backupMode,
							'',
							label,
							actualFile[4],
							str(parent)
						)
												
		return backupSize

	##@brief Wrapped funtion with real functionallity to recover a backup.
	#
	#@param timestamp timestamp of the snapshot to recover 
	#@param presentPath directory being processing
	#@param recoverPath path where place recovered files
	#@param recoverFileList list with files to recover
	# (by default is None that means 'recover all')
	def __recover(self,timestamp, presentPath, recoverPath, recoverFileList = ['/']):
		"""
		Wrapped funtion with real functionallity to recover a backup.
		"""

		if len(recoverFileList) == 1:
			prefix = os.path.dirname(recoverFileList[0]) + '/'
		else:
			prefix = os.path.commonprefix(recoverFileList)
			
		#TODO: Recorrer recoverFileList quitando cualquier archivo que sea hijo
		# de una carpeta que ya esté en recoverFileList.
		
		# For each file we want to recover:
		for path in recoverFileList:
			#TODO: Be sure that this selects the last entry (in timestamp)
			# for each finalpath
			
			# Recover most recent sons for a path if its a directory
			# or the file itself.
			content = self.__db.getSnapshots(str(timestamp),
											path,
											True,
											True,
											'all',
			
											'path')
			
			if content:
				count = len(content)
				filesRecovered = 0
				
				for eachSon in content:
					# Determine if the file was moved
					#moved = self.__db.wasMoved(eachSon[6], timestamp)
					moved = False

					# If the son is not deleted or moved, recover it
					if (eachSon[2] != 'DELETE' and not moved):
						
						# Build up recover directory
						newSnapPath = os.path.join(	self.__SM.getBackupPath(), 
													eachSon[5])					
						candidatePath = eachSon[6]					
						backupFilePath = os.path.join(newSnapPath, candidatePath[1:])
						recoverBackupPath = os.path.join(recoverPath, candidatePath[len(prefix):])
						recoverParentBackupPath = os.path.split(recoverBackupPath)[0]
						
						# To avoid two threads try to make the same directories
						# at the same time.
						self.__lock.acquire()
						try:
							if not os.path.exists(recoverParentBackupPath):
								os.makedirs(recoverParentBackupPath)
						finally:
							self.__lock.release()
					
						# Log info
						self.__log.info("-- Starting recovering file '%s' --" % candidatePath)
						self.__log.info("Building recover file: '%s'" % recoverBackupPath)
						self.__log.info("Looking for '%s.something'" % backupFilePath)
						
						# If it is an integral backup
						if eachSon[3] == 'INT':
							self.__log.info("Restoring INTEGRAL backup.")
							try:
								shutil.copy2(backupFilePath + '.int', recoverBackupPath)
							except IOError:
								self.__log.error(
			"Restoration process failed! Copy '%s' to '%s' did not performed!"
			% (backupFilePath + '.int',recoverBackupPath))
								
						# If it is a differential backup
						elif eachSon[3] == 'DIFF':
							self.__log.info("Restoring DIFFERENTIAL backup.")
							# Look for parent of actual diff
							lastIntegral = self.__db.getByID(eachSon[7])
							
							# If there are no errors
							if lastIntegral:
								# Gets last integral path
								lastIntegralPath = os.path.join(self.__SM.getBackupPath(), 
																lastIntegral[5])
								
								lastIntegralPath = os.path.join(lastIntegralPath,
																candidatePath[1:] + '.int')
								# Recover file
								self.__decodeXDelta3(	lastIntegralPath,
														backupFilePath + '.diff',
														recoverBackupPath)
								# End of restoration
							else:
								# Last integral missed so error
								self.__log.error(
										  "Restoration process failed! "\
										+ "Thera are no integral to restore diff!")
						elif eachSon[3] == 'DIR':
							self.__lock.acquire()
							try:
								if not os.path.exists(recoverBackupPath):
									os.makedirs(recoverBackupPath)
							finally:
								self.__lock.release()
						self.__log.info("-- End of restorating file '%s' --" % candidatePath)
						
						#TODO: Refrescar porcentaje
						
						# FIXME: Delete next 3 lines
						filesRecovered = filesRecovered + 1
						print "%i / %i" % (filesRecovered, count)
		print "recovered"

	##@brief Recover a file or a directory to a specific path at time given.
	#
	# The function behaviour depends on the type of the backup. If it is
	# an integral backup, then the backup itself is copied without .int extension
	# into recover path. Else, if backup is a diff file, then we look for its
	# integral and we apply diff over the integral to recover the file an place it
	# into recover path.
	#
	#@param timestamp timestamp at the file or directory must be recovered.
	#@param originalPathList list of paths for the file(s) or directories
	# to recover.
	#@param recoverPath path where recovered file(s) are placed.
	#@return actual recover path where files(s) has been recovered.
	def recoverBackups(self, snapTimestamp, originalPathList, recoverPath):		
		"""
		Recover a file or a directory to a specific path at time given.
		"""		
		# Get path for timestamp
		snapPath, recoverFolderName = self.__getPath(snapTimestamp)
		
		# Get parent path
		if not len(originalPathList):
			commonPath = ''
		elif len(originalPathList) == 1:
			commonPath = os.path.split(originalPathList[0])[0]
		else:
			commonPath = os.path.commonprefix(originalPathList)[:-1]
		snapPath = os.path.join(snapPath, commonPath[1:])
		
		# Build up recover path
		recoverPath = os.path.abspath(os.path.normpath(recoverPath))
		recoverPath = os.path.normpath(os.path.join(recoverPath, recoverFolderName))		
		if not os.path.exists(recoverPath):
			os.makedirs(recoverPath)
			
		# Recover files
		thread.start_new_thread(
						self.__recover,
						(snapTimestamp,snapPath,recoverPath,originalPathList))

		return recoverPath	

	##@brief Function to delete all backups between two timestamps.
	#
	#@param initialTimestamp Before this timestamp no snapshot will be affected
	#@param endTimestamp After this timestamp no snapshot will be affected
	def deleteBackups(self, path, init_timestamp = 0, final_timestamp = 0):
		"""
		Function to delete all backups between two timestamps.
		"""
		# Set default settings up
		if (init_timestamp == 0 and final_timestamp == 0):
			final_timestamp = int(time.time())
			
		# Hardwork is taken by database that return the new timestamp for moved
		# files, a list of paths to move and a list of path to delete.
		new_timestamp, list_to_move, list_to_delete = \
				self.__db.deleteBackups(path, init_timestamp, final_timestamp)
		
		# To move...
		if new_timestamp and list_to_move:
			for eachFile in list_to_move:
				partial_new_path = os.path.join(self.__getPath(new_timestamp)[0], 
												eachFile[0][1:])
				partial_actual_path = os.path.join(	self.__getPath(eachFile[1])[0],
													eachFile[0][1:])
				if eachFile[2] != 'DIR':
					# Calculate the actual path for the file
					actual_path = os.path.join( self.__getPath(eachFile[1])[0],
												eachFile[0][1:]
												+ "." 
												+ eachFile[2].lower())
					
					# Calculate the new path for the file
					new_path = partial_new_path + "." + eachFile[2].lower()
					
					# Move the file to that path in the filesystem
					os.renames(actual_path, new_path)
				else:
					try:
						os.removedirs(partial_new_path)
					except OSError:
						print "Dir not empty" #FIXME: Replace me by a log
						
		# To delete...
		if list_to_delete:
			# Removes files
			for eachFile in list_to_delete:
				if eachFile[2] != 'DIR' and eachFile[2] != 'NN':
					parentPath = self.__getPath(eachFile[1])[0]
													
					path = os.path.join(parentPath,
										eachFile[0][1:] + "." + eachFile[2].lower())
					
					#Delete the file from the filesystem.
					os.remove(path)
					try:
						os.removedirs(parentPath)
					except OSError:
						print "Dir not empty" #FIXME: Replace me by a log
			# Removes empty files
			for eachFile in list_to_delete:
				if eachFile[2] == 'DIR':
					path = os.path.join(self.__getPath(eachFile[1])[0],
														 eachFile[0][1:])
					
					# Delete the dir from the filesystem.
					try:
						os.removedirs(path)
					except OSError:
						print "Dir not empty" #FIXME: Replace me by a log