# -*- coding: UTF-8 -*-

#Copyright (C) 2008 Adrián Bravo

#This file is part of HDLorean.
#
#HDLorean is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#HDLorean is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with HDLorean.  If not, see <http://www.gnu.org/licenses/>.


import sys
import time
import os


## @brief Watch the status of the battery (if exists).
#
# Few simple methods to know about the battery status.
# @author Adrián
class BatteryMonitor:


	## Battery directory information 
	__battdir="/proc/acpi/battery/"
	## Location of the battery files
	__battinfofile=None
	__battstatefile=None
	__battalarmlimit=None
	## AC status directory
	__acdir="/proc/acpi/ac_adapter/"
	## Location of the AC files
	__acstatefile=None
	## Control some of the possible errors: No acpi module, no laptop...
	__noacpi=False
	__battmaxcap=None


	## @brief Initializes the locations of the files.
	#
	# Search inside the battdir and acdir for the files 
	# needed to read the battery and ac status.
	# @author Adrián
	def __init__(self):
		try:
			batt_contents = os.listdir(self.__battdir)
			ac_contents = os.listdir(self.__acdir)
			# If more or less than one dir inside acdir/battdir we can't determine
			# where the files are.
			if (len(batt_contents) != 1) or (len(ac_contents) != 1):
				if len(batt_contents) != 1:
					print "Could not determine battery directory"
				if len(ac_contents) != 1:
					print "Could not determine ac directory"
				self.__noacpi = True
			else:
				# Usually, the dir contains just one subdir where all
				# the info is stored. However this subdir may vary from
				# one system to another so we list the acdir/battdir and
				# then take the subdir.
				self.__battdir+= batt_contents[0]
				self.__acdir+= ac_contents[0]
				self.__battinfofile = self.__battdir+"/info"
				self.__battstatefile = self.__battdir+"/state"
				self.__battalarmlimit = self.__battdir+"/alarm"
				self.__acstatefile = self.__acdir+"/state"
				self.__battmaxcap = self.__readBattMaxCap()
		except OSError, (errno, strerror):
			print "Error "+str(errno)+": "+strerror
			print "Battery info dirs missing. Probably no ACPI module enabled."
			self.__noacpi = True


	##@brief Determines whether the ac is on or not
	#
	# If acpi module is not enabled or the computer is not a laptop
	# this method always return "on".
	# @return "on" if the ac adapter is plugged in, "off" otherwise. If an error
	# occurs, return None
	# @author Adrián
	def getAcStatus(self):
		if self.__noacpi:
			return "on"
		else:
			try:
				fd=file(self.__acstatefile, "r")
			except IOError:
				return None
			line=fd.readline().split()
			if line[1][0:2]=='on':
				return "on"
			elif line[1][0:3]=="off":
				return "off"
			else:
				return None


	## @brief Internal method to get the maximun battery capacity
	#
	# @return Float with the value of the maximun batt capacity
	# @author Adrián
	def __readBattMaxCap(self):
		try:
			fd=file(self.__battinfofile, "r")
		except IOError:
			battmaxcap = None
			return None
		# Skip lines until we reach the Capacity info
		present = fd.readline()
		if present.split()[1] != "yes":
			print "WARNING: Battery reported as not present!"
			self.__noacpi = True
			return None
		fd.readline()
		battmaxcap=float(fd.readline().split()[3])
		fd.close()
		return battmaxcap


	## @brief Returns the maximun battery capacity
	#
	# @return Float with the maximun batt capacity. If an error occurs
	# return None
	# @author Adrián
	def getBattMaxCap(self):
		return self.__battmaxcap

	
	## @brief Gets the remaining battery charge percent
	#
	# If the acpi module is not enabled or the info files about battery
	# were missing, this method always returns fully charged (100).
	# @return Int with the percentual value of the remaining battery. If an error
	# occurs it return None
	# @author Adrián
	def getRemainingCharge(self):
		if self.__noacpi:
			return 100
		else:
			try:
				fd=file(self.__battstatefile, "r")
			except IOError:
				return None
			present=fd.readline()
			if present.split()[1] != "yes":
				print "WARNING: Battery reported as not present!"
				return None
			# Skip needless info
			fd.readline()
			fd.readline()
			fd.readline()
			return int(float(fd.readline().split()[2])/self.__battmaxcap*100)
