# -*- coding: UTF-8 -*-

#Copyright (C) 2008 Adrián Bravo.

#This file is part of HDLorean.
#
#HDLorean is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#HDLorean is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with HDLorean.  If not, see <http://www.gnu.org/licenses/>.


import os
import copy
import string
import gobject
import fnmatch
import GlobalExceptions
from hdlogger import HDLogger
# Suported special chars by fnmatch
# Pattern   	Meaning 
# * 		matches everything
# ? 		matches any single character
# [seq] 	matches any character in seq
# [!seq] 	matches any character not in seq

# Built-in class that provides methods to create and manage a
# configuration file with a predefined syntax.
import ConfigParser

## @brief Manage config parameters and file.
#
# Wrapper around ConfigParser to provide hdlorean a simple interface with
# the configuration provided by the user and the program defaults. It also
# makes some operations on the config to make hdlorean easier manage of the
# config.
# @author Adrián
class ConfigFileManager(object):

	## Files to be watched and backuped
	__watchedList = set()
	## Path to the config file
	__configFilePath= ''
	## ConfigParser instance, to manage the file.
	__config = ConfigParser.SafeConfigParser()
	# Sets avoid repeated elements
	## Patterns lists
	__reBlacklist = set()
	__reWhitelist = set()
	## Paths lists
	__blacklist = set()
	__whitelist = set()
	## A set containing the paths of the folders that have
	# been expanded and so removed from the whitelist/watchedlist.
	# Its used by inotify that needs to distinguish between them.
	__parentsList = set()
	# Needed to implement Singleton pattern
	__instance = None
	## This param determines if the findFiles() method has been called
	# at least once.
	__initialized = False
	##
	__logger = None
	##
	__log = None
	## Instance of watcher
	__watcher = None


	## @brief Initializes some parts of the object CFM.
	#
	# Initializes dbus instance an guarantees that only one is created.
	# Checks if the config file exists and if so, read it. Otherwise create
	# it with default settings.
	# @author Adrián, Jorge, Danieloop
	def __init__(self,watcher=None):
		if os.path.exists(self.__configFilePath):
			self.__config.read(self.__configFilePath)
		else:
			self.__log.info("Config file is missing. It will be regenerated with the default values")
			self.setDefaultConfig()
		snapdir = self.getBackupDirectory()
		if not os.path.exists(snapdir):
				os.mkdir(snapdir)


	# @brief Initialize the instance. Guarantees just one instance of this class.
	#
	# @param A watcher instance to talk to him
	# @author Adrián
	def __new__(self,watcher=None):
		# if the class has not been instanciated before do it
		# else return instance.
		if self.__instance is None:
			self.__configFilePath = os.getenv("HOME")+"/.hdlorean/config.cfg"
			self.__logger = HDLogger.HDLogger()
			self.__log = self.__logger.setupHDLogger('hdlorean.Watcher.ConfigFileManager')
			self.__instance = object.__new__(self)
		
		if watcher is not None:
			self.__watcher = watcher
		return self.__instance

	
	## @brief Starts the search of the files to be watched.
	#
	# If the files have not been searched yet, finds them and
	# put the correct values in all the lists (black, white, watched...)
	# @return True if the files are searched, False if they were previously found.
	# @author Adrián
	def findFiles(self):
		#TODO: FIXME: This method should be thread safe... 
		if self.__initialized == False:
			self.__mergeFileLists(self.__getOptionValue('Lists','start_dir'))
			self.__initialized = True
			return True
		else:
			return False
	
	
	#*********************DEPRECATED*****************************************************************
	##@brief implements behavior for config file updates. Deprecated
	#
	# When config file is modified the sets of items must be reloaded
	# @return  tuple containing the differences between the old (white|parent)list and
	# the new (white|parent)list
	# @author Jorge
	def updateFiles(self):
		#first of all, copy the "old" list. We need some of them to generate the differences with the
		#new ones
		oldWatchedList = self.__watchedList.copy()
		oldParents = self.__parentsList.copy()
		#config file read is forced
		self.__config.read(self.__configFilePath)
		#generate the new sets
		self.__mergeFileLists(self.__getOptionValue('Lists','start_dir'))
		#obtain the new elements of watchedlist
		toAddInWatched = self.__watchedList.difference(oldWatchedList)
		#obtain the elements that are no longer in watchedlist
		toRemoveFromWatched = oldWatchedList.difference(self.__watchedList)
		# same as above with parent list
		toAddInParents = self.__parentsList.difference(oldParents)
		toRemoveFromParents = oldParents.difference(self.__parentsList)
		#return a tuple containing the previous sets
		return toAddInWatched,toRemoveFromWatched,toAddInParents,toRemoveFromParents
		#TODO implement a removeFromParent in PyinotifyHandler
	

	## @brief Obtains the list of files/dirs to be watched/backed up.
	#
	# Unify both whitelist and blacklist (of patterns and of paths) in a 
	# new list containing only the files that must be watched/backuped 
	# by hdlorean. The method includes pattern expansion provided by fnmatch.
	# The list is saved into watchedList
	# @param start_dir The root path where the search of patterns will start.
	# @author Adrián,Jorge
	def __mergeFileLists(self,start_dir):
		# Prepare a tuple with the blacklisted-patterns and the whitelisted ones
		# to walk trough the dirtree.
		#FIXME: Should the section name for the black/white-list be
		# hardcoded here? Recovering their values will be more efficient,
		# though it will be less general.
		re_blacklst_parts = self.__getOptionValue('Lists','re_blacklist').split(";")
		re_whitelst_parts = self.__getOptionValue('Lists','re_whitelist').split(";")
		regex = (re_blacklst_parts,re_whitelst_parts)
		# Update the contents of sets with the lists in the config file.	
		self.__blacklist.update(self.__getOptionValue('Lists','blacklist').split(";"))
		#self.__blacklist=self.__blacklist.intersection(self.__getOptionValue('Lists','blacklist').split(";"))
		self.__whitelist.update(self.__getOptionValue('Lists','whitelist').split(";"))
		#self.__whitelist=self.__whitelist.intersection(self.__getOptionValue('Lists','whitelist').split(";"))
		#This is ugly but only one module uses parentList and this serves him well
		#self.__parentsList = set()
		
		# Sets cannot change during iteration, so we need an aux set.
		whitelist_aux = self.__whitelist.copy()
		# The algorithm is as follows:
		# For each path in whitelist (wpath), find every path in blacklist (bpath) that contains wpath
		# as substring. Then adds to whitelist all the contents of the dir wpath (just one level below)
		# except the ones that was substring. After that, remove wpath from the list.
		for wpath in self.__whitelist:
			# Tuned filter -> Comprehension list. Gets all the paths in blacklist that have
			# wpath as substring.
			substr_list = []
			substr_list.extend([bpath for bpath in self.__blacklist if self.__substr(bpath,wpath)])
			if substr_list != []:
				try:
					# Adds all the contents of wpath except substr_list and extract wpath
					# Gets all the contents of wpath directory.
					path_listing = os.listdir(wpath)
					# Change path_listing to a set containing absolute paths for the contents of wpath.
					path_listing = set([os.path.join(wpath,relpath) for relpath in path_listing if relpath])
					# Puts in aux_set the contents of wpath that are not in substr_list
					aux_set = path_listing.difference(set(substr_list))
					# Adds to whitelist_aux the elements of aux_set, then drop wpath.
					whitelist_aux.update(aux_set)
					whitelist_aux.discard(wpath)
					self.__parentsList.add(wpath)
				except OSError, (errno,errmsg):
					self.__log.error(repr(errno) + ':' + errmsg)

		self.__whitelist = whitelist_aux
		# Gets the concrete (expanded) files and dirs matching the patterns.
		# If there is some pattern to find
		if ((regex[0] != ['']) and (regex[0] != [])) or ((regex[1] != ['']) and (regex[1] != [])):
			os.path.walk(start_dir, self.__visitor,regex)
		# Delete from pattern whitelist the items of the pattern blacklist
		self.__reWhitelist.difference_update(self.__reBlacklist)
		# Now, having the concrete patterns and the concrete paths, we merge both whitelists in the
		# watchedList.
		# Add the whitelisted-patterns to the watchedList, then remove the blacklisted ones.
		self.__whitelist.update(self.__reWhitelist)
		self.__whitelist.difference_update(self.__reBlacklist)
		self.__watchedList = self.__whitelist


	## @brief Determines if wpath is initial substr of bpath.
	#
	# Uses find to determine if the position in bpath of wpath is 0
	# and then return 1 (true). Otherwise it returns 0 (false).
	# @param bpath String to test if it has wpath as initial substr.
	# @param wpath String to test if is initial substr of bpath
	# @return 1 if wpath is substr of bpath. 0 Otherwise.
	# @author Adrián
	def __substr(self,bpath,wpath):
		if bpath.find(wpath) == 0:
			return 1
		else:
			return 0	

	#
	# Internal function to be used in conjunction with os.path.walk.
	# Each time this function is invoked, finds matches
	# between the regexmatrix patterns and the list of files
	# contained in the filelist arg. Returns each match file in
	# the corresponding list (whitelist or blacklist).
	# @param regexmatrix A matrix containing in the first row the
	# patterns of the blacklist and in the second row the patterns
	# of the whitelist. This argument is provided to os.path.walk
	# by the caller.
	# @param dirname The current dir. Provided by os.path.walk.
	# @param filelist All the files contained in dirname. Provided by
	# os.path.walk
	# @see os.path.walk
	# @see
	# @author Adrián
	# TODO: Don't match regex against dirs?. Avoid repeated patterns between blacklist and whitelist
	def __visitor(self,regexmatrix,dirname,filelist):
		# Walks through the regexmatrix matching all the files in the
		# current dir with all the patterns in the matrix.
		# If a match is found, then it's added to either the black or white
		# list, assuming the first row (0) in regexmatrix contains the
		# blacklist patterns and the second row contais the whitelist ones.
		
		# If the current dir is in the blacklist we shouldn't find patterns there. But
		# indeed, we need the reBlackList complete to check against it, so we find the patterns.
		# If it's in the whitelist we'll add the whole dir, so we won't find patterns there. We
		# control this with addtowhitelist
		if dirname in self.__whitelist:
			addtowhitelist = True
		else:
			addtowhitelist = False

		for row in range(len(regexmatrix)):
			for regex in regexmatrix[row]:
				if ((row == 0) or (row == 1)) and addtowhitelist:
					# Split the regex string in something with the form *.py or /etc/*.py with rpartition.
					# If regex is "pure" we must match anyway, otherwise split pattern by / and if first 
					# component is substr of dirname, we match.
					# 
					# Important: Its necesary to re-initialize listofmatches or it will keep last time value
					# and add redundant information to the lists
					listofmatches = []
					regex_parts = []
					# Emulate rpartition with find and slice because rpartition is a python 2.5 feature
					pos = regex.rfind("/")
					if pos != -1:
						regex_parts.append(regex[:pos])
						regex_parts.append(regex[pos])
						regex_parts.append(regex[pos+1:])
					else:
						regex_parts.append('')
						regex_parts.append('')
						regex_parts.append(regex)
					# If the first element of tuple is empty, it means that regex doesn't contain a path part, so its
					# "pure regex", something like (*.py)
					if regex_parts[0] == '':
						listofmatches = fnmatch.filter(filelist,regex)
					# If it's not empty, the regex contains a prefix marking the root path where the regex should be matched.
					else:
						# If first part of regex (the path part) is substring of the actual directory
						# then we must check for coincidences with the final part of the regex (the *.py part).
						if dirname.count(regex_parts[0]) != 0:
							listofmatches = fnmatch.filter(filelist,regex_parts[2])
					# If any match is found, it's added to the blacklist or whitelist based on the row of the regex
					# in regexmatrix. Row 0 = blacklist, Row 1 = whitelist.
					if listofmatches != []:
						#TODO: THIS IS EVERYTHING BUT EFFICIENT!! Might be made inside filter (reimplement?)
						# or reimplementing os.path.walk so it really steps into the dir.
						for i in range(len(listofmatches)):
							listofmatches[i] = os.path.join(dirname,listofmatches[i])
							if row == 0:
								self.__reBlacklist.add(listofmatches[i])
							else:
								self.__reWhitelist.add(listofmatches[i])	


	## @brief Writes the Config File Object to the File itself.
	#
	# Write all the changes made in the dynamic config file object
	# to the disk.
	# @author Adrián
	def __writeConfigObject(self):
		file = open(self.__configFilePath,"w")
		self.__config.write(file)

	
	## @brief Gets the list of "must be watched" files.
	# 
	# @see __mergeFileList
	# @return The list of files that must be watched/backup-ed
	# @author Adrián
	def getWatchedFiles(self):
		return self.__watchedList	
	
	
	## @brief Overwrite the contents of the lists and the config file and forces new merge.
	#
	# Set the contents of the lists to the user defined ones, and updates the config file with them.
	# Then it forces a new merge of the lists and a search of the patterns. It expects a semicolon separated string.
	# @param whitelist New string path1;path2;path3 of files the user want to watch
	# @param reWhitelist New string path1;path2;path3 of patterns the user want to watch
	# @param blacklist New string path1;path2;path3 of files the user doesnt want to watch
	# @param reBlacklist New string path1;path2;path3 of patterns the user doesnt want to watch.
	# @see findFiles
	# @author Adrián
	def setFileLists(self, whitelist, reWhitelist, blacklist, reBlacklist):
		# overwrite the contents of the lists with the new config
		# given by the user
		if blacklist is not None:
			#blacklist = str(blacklist.encode('utf-8','strict'))
			self.__addConfigItem("Lists","blacklist",blacklist)
			self.__blacklist = set(blacklist.split(";"))
		if whitelist is not None:
			#whitelist = str(whitelist.encode('utf-8','strict'))
			self.__addConfigItem("Lists","whitelist",whitelist)
			self.__whitelist = set(whitelist.split(";"))
		if reWhitelist is not None:
			#reWhitelist = str(reWhitelist.encode('utf-8','strict'))
			self.__addConfigItem("Lists","re_whitelist",reWhitelist)
			self.__reWhitelist = set(reWhitelist.split(";"))
		if reBlacklist is not None:
			#reBlacklist = str(reBlacklist.encode('utf-8','strict'))
			self.__addConfigItem("Lists","re_blacklist",reBlacklist)
			self.__reBlacklist = set(reBlacklist.split(";"))
		# Then force a new search of patterns and a new merge of lists
		self.__initialized = False
		self.findFiles()
		# Notify watcher of changes on the lists
		self.__watcher.updateLists()


	##
	#
	#
	def addToPblacklist(self,items):
		self.__blacklist.update(set(items))
		
		

	## @brief Gets the Procesed patterns whitelist
	# 
	# The state of the patterns whitelist changes along
	# the execution of this class, be carefull when you
	# get the list.
	# @return A Set containing the whitelisted patterns
	def getPReWhiteList(self):
		return self.__reWhitelist


	## @brief Gets the Procesed patterns blacklist
	# 
	# The state of the patterns blacklist changes along
	# the execution of this class, be carefull when you
	# get the list.
	# @return A Set containing the blacklisted patterns
	def getPReBlackList(self):
		return self.__reBlacklist


	## @brief Gets the Procesed paths whitelist
	#
	# Keep in mind that the whitelist is not the filtered
	# list of the files that should be watched. It's an
	# aux structure used to obtain the final watchedList.
	# @see getWatchedList
	# @return A Set containing the paths whitelist.
	def getPWhiteList(self):
		return self.__whitelist


	## @brief Gets the Procesed Paths blacklist
	#
	# @see getReBlacklist
	# @return A Set containing the paths blacklisted
	def getPBlackList(self):
		return self.__blacklist

	
	## @brief Gets the Parents List
	#
	# Returns the list of the paths that have been expanded
	# and removed from the whitelist, so Inotify can add them
	# no recursively.
	# @see parentsList
	# @return A Set containing the expanded paths
	# @author Adrián
	def getParentsList(self):
		return self.__parentsList


	## @brief Obtains a list of the options in the section especified
	#
	# Returns in a list the options defined in the section especified.
	# @param section The section to retrieve the options from
	# @return A list filled with the options names contained in the section.
	# @author Adrián
	def getOptions(self,section):
		return self.__config.options(section)
	

	## @brief Obtains a list of the sections in the config file.
	#
	# Returns in a list the sections available in the config file.
	# @return A list filled with the sections names contained in the file.
	# @author Adrián
	def getSections(self):
		return self.__config.sections()


	## @brief Determines if the section given has te option especified
	#
	# Search in the given section for the option passed as argument and
	# return a boolean.
	# @param section The section that may contain de option.
	# @param option The searched option.
	# @return True if the option exists whithin the given section and 
	# False Otherwise
	# @author Adrián
	def hasOption(self,section,option):
		return self.__config.has_option(section,option)

	
	## @brief Determines if the section given exists in the config file.
	#
	# Search in the config file for the section passed as argument and
	# return a boolean.
	# @param section The section the caller wants to know if exists.
	# @return True if the section exists and False Otherwise
	# @author Adrián
	def hasSection(self,section):
		return self.__config.has_section(section)


	## @brief: Obtains the value of the given option in the given section.
	#
	# If the caller needs the option in float, int or bool, it can
	# call the function with that arg in the last place (type).
	# Supported types are float, int and bool. Call the function
	# with just three arguments to get the option without type changes.
	# @param section Name of the section which contains the option.
	# @param option Name of the option which value will be returned.
	# @param type The type of the option. It can be ommited.
	# @return The value of the especified option with the type especified.
	# @author Adriánst
	def __getOptionValue(self,section,option,type="None"):
		# The methods getXXX coerces the option in the especified section
		# to that type XXXX.
		if type == "int":
			return self.__config.getint(section,option)
		elif type == "float":
			return self.__config.getfloat(section,option)
		elif type == "bool":
			return self.__config.getboolean(section,option)
		else:
			return self.__config.get(section,option)


	## @brief Obtains the value of the given option if exists whithin any section.
	#
	# Moves through all the sections in the Config File looking for the given option name.
	# If it exists somewhere, its value is returned. Otherwise a NoOptionError exception
	# is raised. If the name of the section is known, __getOptionValue should be used instead
	# of this method, because its much more efficient.
	# @param option Name of the option which value will be returned.
	# @param type The type of the option. It can be ommited.
	# @return The value of the especified option with the type especified.
	# @throw NoOptionError The searched option doesn't exist.
	# @see __getOptionValue
	# @author Adrián
	# REMEMBER: This function is DEPRECATED and it probably wont work as you expect
	def __getFileOptionValue(self,option):
		for section in self.getSections():
			if self.hasOption(section,option):
				return self.__getOptionValue(section,option)
			else:
				self.__log.warning('Option not found')
				raise GlobalExceptions.OptionNotFound

	
	## @brief Add new option to the config file
	#
	# This method tries to add a new option with its value to the section especified.
	# If the section doesn't exist raise an exception. Otherwise it adds the option to
	# the dynamic estructures and to the config file.
	# @param section The name of the section in which the option will be added
	# @param option The name of the option that will be added
	# @param value The value of the new option added
	# @throw NoSectionError The especified section does not exist
	# @throw TypeError Value must be a string
	# @author Adrián
	def __addConfigItem(self,section,option,value):
		self.__config.set(section,self.__config.optionxform(option),value)
		self.__writeConfigObject()


	## @brief Delete a config param from the config file.
	#
	# Remove the option especified from the section given in
	# the config file (dynamic and static).
	# @param section The section containing the option
	# @param option The option to be deleted
	# @throw NoSectionError The section especified doesn't exist.
	# @author Adrián
	def removeConfigItem(self,section,option):
		self.__config.remove_option(section,option)
		self.__writeConfigObject()


	## @brief Delete a section from de config file
	#
	# Remove a section of the config file if it exists.
	# @param The section to remove
	# @return True if the section really exists and went deleted properly.
	# False otherwise.
	# @author Adrián
	def removeSection(self,section):
		result = self.__config.remove_section(section)
		self.__writeConfigObject()
		return result


	## @brief Add new section to the config file
	#
	# This method add a new section to the dynamic structures and
	# writes it to the config file..
	# @param section The name of the section that will be added
	# @throw DuplicateSectionError The section especified already exists!
	# @author Adrián
	def addSection(self,section):
		self.__config.add_section(section)
		self.__writeConfigObject()
	

	## @brief Getter of initialized, determines if the findFiles has been
	# invoked at least once.
	#
	# @return Boolean. True if the files have been initialized, false otherwise.
	# @author Adrián
	def isInitialized(self):
		return self.__initialized

	
	## @brief Getter of the option NautilusIntegration.
	#
	# @return Boolean. True if nautilus integration is activated, false otherwise.
	# @author Adrián
	def getNautilusIntegration(self):
		return self.__getOptionValue("General","nautilusintegration","bool")

	
	## @brief Setter of the option NautilusIntegration.
	#
	# @param nautint. True to activate integration, false otherwise.
	# @author Adrián
	def setNautilusIntegration(self,nautint):
		self.__addConfigItem("General","nautilusintegration",nautint)


	## @brief Getter of the option Periodically, determines the frequency to do backups.
	#
	# @return String. The frequency.
	# @author Adrián
	def getPeriodically(self):
		return self.__getOptionValue("BackupInterval","periodically")
	
	
	## @brief Setter of the option Periodically, determines the frequency to do backups.
	#
	# @param period. The frequency will be used.
	# @author Adrián
	def setPeriodically(self,period):
		self.__addConfigItem("BackupInterval","periodically",period)
		self.__watcher.updateCrontab()

	
	## @brief Getter of the option EveryChange, determines if inotify have to work.
	#
	# @return Boolean. True if inotify is activated, false otherwise.
	# @author Adrián
	def getEveryChange(self):
		return self.__getOptionValue("BackupInterval","everychange","bool")
	
	
	## @brief Setter of the option EveryChange, determines if inotify have to work.
	#
	# @param every. True to activate inotify, false otherwise.
	# @author Adrián
	def setEveryChange(self,every):
		self.__addConfigItem("BackupInterval","everychange",every)


	## @brief Getter of the option DeleteIfSpace, determines the maximum space (in MB)
	# what can be used by backups.
	#
	# @return int. The maximum space.
	# @author Adrián
	def getDeleteIfSpace(self):
		return self.__getOptionValue("HDUsage","deleteifspace","int")
	
	
	## @brief Setter of the option DeleteIfSpace, determines the maximum space (in MB)
	# what can be used by backups.
	#
	# @param delifspace. The maximum space.
	# @author Adrián
	def setDeleteIfSpace(self,delifspace):
		self.__addConfigItem("HDUsage","deleteifspace",delifspace)

	
	## @brief Getter of the option DeleteIfOlder, if it's necessary to free space, it determines
	# the age from files have to be deleted.
	#
	# @return int. The age.
	# @author Adrián
	def getDeleteIfOlder(self):
		return self.__getOptionValue("HDUsage","deleteifolder","int")
	

	## @brief Setter of the option DeleteIfOlder, if it's necessary to free space, it determines
	# the age from files have to be deleted.
	#
	# @return int. The age.
	# @author Adrián
	def setDeleteIfOlder(self,delifolder):
		self.__addConfigItem("HDUsage","deleteifolder",delifolder)
	
	
	## @brief Getter of the option BackupDirectory, determines the path where will be saved the backups.
	#
	# @return String. The path of the backup directory.
	# @author Adrián
	def getBackupDirectory(self):
		return self.__getOptionValue("HDUsage","backupdirectory")
	
	
	## @brief Setter of the option BackupDirectory, determines the path where will be saved the backups.
	#
	# @param backupdir. The path of the backup directory.
	# @author Adrián
	def setBackupDirectory(self,backupdir):
		self.__addConfigItem("HDUsage","backupdirectory",backupdir)
	
	
	## @brief Getter of the option StartDir, determines the directory from where patterns are searched.
	#
	# @return String. The path of the start directory.
	# @author Adrián
	def getStartDir(self):
		return self.__getOptionValue("Lists","start_dir")
	

	## @brief Setter of the option StartDir, determines the directory from where patterns are searched.
	#
	# @param startdir. The path of the start directory.
	# @author Adrián
	def setStartDir(self,startdir):
		self.__addConfigItem("Lists","start_dir",startdir)
		# Then force a new search of patterns and a new merge of lists
		self.__initialized = False
		self.findFiles()
		# Notify watcher of changes on the lists
		self.__watcher.updateLists()


	## @brief Getter of the option BlackList, determines the excluded directories.
	#
	# @return Array of strings. The paths of the excluded directories.
	# @author Adrián
	def getBlackList(self):
		return self.__getOptionValue("Lists","blacklist")
	
	
	## @brief Setter of the option BlackList, determines the excluded directories.
	#
	# @param blacklist. List of paths to be excluded.
	# @author Adrián
	def setBlackList(self,blacklist):
		self.setFileLists(None,None,blacklist,None)
		# Then force a new search of patterns and a new merge of lists
		self.__initialized = False
		self.findFiles()
		# Notify watcher of changes on the lists
		self.__watcher.updateLists()


	## @brief Getter of the option WhiteList, determines the watched directories.
	#
	# @return Array of strings. The paths of the watched directories.
	# @author Adrián
	def getWhiteList(self):
		return self.__getOptionValue("Lists","whitelist")
	
	
	## @brief Setter of the option WhiteList, determines the watched directories.
	#
	# @param whitelist. The paths of the watched directories.
	# @author Adrián
	def setWhiteList(self,whitelist):
		self.setFileLists(whitelist,None,None,None)
		# Then force a new search of patterns and a new merge of lists
		self.__initialized = False
		self.findFiles()
		# Notify watcher of changes on the lists
		self.__watcher.updateLists()

	##
	#
	#
	def getReWhiteList(self):
		return self.__getOptionValue("Lists","re_whitelist")
	
	
	##
	#
	#
	def setReWhiteList(self,rewhitelist):
		self.setFileLists(None,rewhitelist,None,None)
		# Then force a new search of patterns and a new merge of lists
		self.__initialized = False
		self.findFiles()
		# Notify watcher of changes on the lists
		self.__watcher.updateLists()


	##
	#
	#
	def getReBlackList(self):
		return self.__getOptionValue("Lists","re_blacklist")
	

	##
	#
	#
	def addToPReBlacklist(self,items):
		self.__blacklist.update(set(items))


	##
	#
	#
	def setReBlackList(self,reblacklist):
		self.setFileLists(None,None,None,reblacklist)
		# Then force a new search of patterns and a new merge of lists
		self.__initialized = False
		self.findFiles()
		# Notify watcher of changes on the lists
		self.__watcher.updateLists()


	## @brief Getter of the option BatteryChargeLimit, determines the minimal percentage of battery
	# to make snapshots.
	#
	# @return battchargelimit battery limit assigned.
	# @author Adrián
	def getBatteryChargeLimit(self):
		return self.__getOptionValue("Limits","batterychargelimit","int")
	
	
	## @brief Setter of the option BatteryChargeLimit, determines the minimal percentage of battery
	# to make snapshots.
	#
	# @param battchargelimit battery limit assigned.
	# @author Adrián
	def setBatteryChargeLimit(self,battchargelimit):
		self.__addConfigItem("Limits","batterychargelimit",battchargelimit)

		
	## @brief Getter of the option CpuUseLimit, determines the maximum percentage of cpu
	# to make snapshots.
	#
	# @return limit of cpu assigned.
	# @author Adrián
	def getCpuUseLimit(self):
		return self.__getOptionValue("Limits","cpuuselimit","int")

	
	## @brief Setter of the option CpuUseLimit, determines the maximum percentage of cpu
	# to make snapshots.
	# @param cpuuselim limit of cpu assigned.
	# @author Adrián
	def setCpuUseLimit(self,cpuuselim):
		self.__addConfigItem("Limits","cpuuselimit",cpuuselim)


	## @brief Getter of the option MaxBackupSize, determines the maximum size (in MB)
	# of the backup directory.
	#
	# @return maximum size of backup directory.
	# @author Adrián
	def getMaxBackupSize(self):
		return self.__getOptionValue("Limits","maxbackupsize","int")
	

	## @brief Setter of the option MaxBackupSize, determines the maximum size (in MB)
	# of the backup directory.
	#
	# @param maximum size of backup directory.
	# @author Adrián
	def setMaxBackupSize(self,maxbackupsize):
		self.__addConfigItem("Limits","maxbackupsize",maxbackupsize)

	
	## @brief Create the default config file with default settings
	#
	# Check if the hdlorean dir exists and creates it if it doesnt. Then
	# creates the config file and write the defaults to it.
	# @author Adrián
	def setDefaultConfig(self):
		if not os.path.exists(self.__configFilePath):
			home = os.getenv("HOME")
			if not os.path.exists(home+"/.hdlorean"):
				os.mkdir(home+"/.hdlorean")
				#shutil.copy2()		#FIXME?? Copy the dbusmessage.py??
			os.mknod(self.__configFilePath)

		#Initialize a configparser object.
		self.__config.read(self.__configFilePath)
		#Add the default sections
		self.addSection("General")
		self.addSection("BackupInterval")
		self.addSection("Lists")
		self.addSection("HDUsage")
		self.addSection("Limits")
		#Add the default params
		self.__addConfigItem("General","nautilusintegration", "False")
		self.__addConfigItem("BackupInterval","periodically","1 hour")
		self.__addConfigItem("BackupInterval","everychange","True")
		self.__addConfigItem("HDUsage","deleteifspace","0")
		self.__addConfigItem("HDUsage","deleteifolder","0")
		self.__addConfigItem("HDUsage","backupdirectory",os.getenv('HOME')+'/.hdlorean/snapshots')
		self.__addConfigItem("Lists","start_dir",home)
		self.__addConfigItem("Lists","re_blacklist","*/.*")
		self.__addConfigItem("Lists","whitelist",home)
		self.__addConfigItem("Lists","re_whitelist","")
		self.__addConfigItem("Lists","blacklist","")
		self.__addConfigItem("Limits","batterychargelimit","20")
		self.__addConfigItem("Limits","cpuuselimit","0")
		self.__addConfigItem("Limits","maxbackupsize","0")

		self.__writeConfigObject()
