# -*- coding: UTF-8 -*-

#Copyright (C) 2008 Adrián Bravo, Diana Lopez

#This file is part of HDLorean.
#
#HDLorean is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#HDLorean is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with HDLorean.  If not, see <http://www.gnu.org/licenses/>.


#python modules
import string
import time
import os
import signal
import tempfile
import ConfigFileManager
import GlobalExceptions
from hdlogger import HDLogger
from snapshotManager import SnapshotManager


##
#
# @author Diana, Adrián
class CronHandler:
	__cfm = None
	__crontabPath = "/usr/bin/crontab"
	__cronCommand = ""
	__lastBackupF=""
	__logger = None
	__log = None
	__snapshotManager = None
	__watcher = None


	##
	#
	# @author Adrián, Diana
	def __init__(self,watcher):
		self.__logger = HDLogger.HDLogger()
		self.__cronCommand = os.getenv("PWD")+'/watcher/CronCommand.py'
		self.__watcher = watcher
		self.__log = self.__logger.setupHDLogger('hdlorean.Watcher.CronHandler')
		self.__cfm = ConfigFileManager.ConfigFileManager()
		self.__snapshotManager = SnapshotManager.SnapshotManager()
		# prepare hdlorean to recieve sigusr1 signal
		signal.signal(signal.SIGUSR1,self.__cronHandler)
		# Check if its later than the time in the crontab
		# FIXME: Uncomment when the function is available
		#lastCronBackup = self.__snapshotManager.getLastCronBackup()
		acttime = time.time()
		lastCronBackup = acttime + 2
		period = self.__cfm.getPeriodically()
		if (acttime-lastCronBackup>period):
			# Make new cron snapshot
			self.__watcher.snapshotRequest("CRON")


	## @brief Request a new snapshot by CRON 
	#
	# @author Adrián
	def __cronHandler(self,signum,stackframe):
		self.__watcher.snapshotRequest("CRON")


	## @brief Adds a task to the crontab.
	#
	# Creates a temporal file with old tasks and new ones and then updates crontab.
	# Only adds if task isn't in the crontab yet.
	# @author Diana
	def addTask (self):
		if ((self.__cfm.hasSection("BackupInterval")) and (self.__cfm.hasOption("BackupInterval","Periodically"))):
			ta = self.__cfm.getPeriodically()
			if (ta=='None'):
				self.removeTask()
			else:
				task = self.__convertTask(ta)
				execute = self.__crontabPath + " -l"
				lines = os.popen(execute).readlines()
				#Creates a temporary file in the most secure manner possible.
				tmpfile = tempfile.mkstemp ()
				fd, path = tmpfile
				# Returns an open file object connected to the file decronCommandor fd to write.
				tmp = os.fdopen(fd, 'w')
				count = 0
				exist = False
				for line in lines:
				## Ignore the first three comments:
		
				# DO NOT EDIT THIS FILE - edit the master and reinstall.
				## (/tmp/crontab.XXXXXX installed on Xxx Xxx  x xx:xx:xx xxxx)
				# (Cron version -- $Id: crontab.c,v 2.13 1994/01/17 03:20:37 vixie Exp $)
					if not (count < 3 and len(line) > 1 and line[0] == "#"):
						if (line.endswith("#hdlorean\n")):
							#It already exists another hdlorean task
							#remove it and add new one
							#if task is the same not delete
							if not self.__compareTask(ta,line):
								#if tasks are different, not write it
								exist = False
							else:
								# if tasks are the same, write it
								exist = True
								tmp.write (line)
								if line[len(line)-1] != '\n':
									tmp.write ("\n")
					else:		
							tmp.write (line)
							if line[len(line)-1] != '\n':
								tmp.write ("\n")
					count = count + 1
				if not exist:
					# Writes the task at the end.
					tmp.write(task+self.__cronCommand+"	#hdlorean\n")
					tmp.close ()
					# Replace crontab config with new one in file.
					os.system (self.__crontabPath + " " + path)
				else:
					tmp.close ()
		else:
			raise GlobalExceptions.OptionNotFound


	## @brief Removes a task to the crontab if it exists.
	#
	# Creates a temporal file with old tasks but the one to remove and then updates crontab.
	# Only removes if task exist in the crontab. If it doesn't, the crontab doesn't change.
	# @author Diana
	def removeTask (self):
		execute = self.__crontabPath + " -l"
		lines = os.popen(execute).readlines()
		#Creates a temporary file in the most secure manner possible.
		tmpfile = tempfile.mkstemp ()
		fd, path = tmpfile
		# Returns an open file object connected to the file decronCommandor fd to write.
		tmp = os.fdopen(fd, 'w')
		count = 0
		exist = False
		for line in lines:
		## Ignore the first three comments:
	
		## DO NOT EDIT THIS FILE - edit the master and reinstall.
		## (/tmp/crontab.XXXXXX installed on Xxx Xxx  x xx:xx:xx xxxx)
		## (Cron version -- $Id: crontab.py 1000 2007-12-02 18:08:02Z gauteh $)

			if not (count < 3 and len(line) > 1 and line[0] == "#"):
				if (line.endswith("#hdlorean\n")):
					exist = True
				else:
					tmp.write (line)
					if line[len(line)-1] != '\n':
						tmp.write ("\n")
			count = count + 1
		tmp.close ()
		if exist:
			# Replace crontab config with new one in file.
			os.system (self.__crontabPath + " " + path)


	## @brief Takes the period given by user form the config file and converts it to crontab syntax.
	#
	#     m h dom mon dow user    command
	#     m: minutes
	#     h: hour
	#     dom: day of the month 
	#     mon: month
	#     dow: day of the week
	#     user: user who executes the command
	#     command: absolut path of the cronCommand to execute
	#     Example: 5 10 * * * /home/user/user_cronCommand
	#     every day at 10.05 am user_cronCommand is executed.
	# @param t task from file config (string)
	# @author Diana
	def __convertTask (self,t):
		if t == "1 min":
			return "* * * * *    "
		if t == "5 min":
			return "*/5 * * * *    "
		if t == "10 min":
			return "*/10 * * * *    "
		if t == "30 min":
			return "*/30 * * * *    "
		if t == "1 hour":
			return "0 */1 * * *    "
		if t == "1 day":
			fecha = time.localtime()
			minute=fecha[4]
			hour=fecha [3]
			st=str(minute)+" "+str(hour)+" */1 * *	"
			return st
	
	
	## @brief Takes the period given and the period in the crontab and checks if are the same
	#.
	# @param t (string)    m h dom mon dow user    command
	#     m: minutes
	#     h: hour
	#     dom: day of the month 
	#     mon: month
	#     dow: day of the week
	#     user: user who executes the command
	#     command: absolut path of the cronCommand to execute
	#     Example: 5 10 * * * /home/user/user_cronCommand
	#     every day at 10.05 am user_cronCommand is executed.
	# @param t task from file config (string)
	# @param l task from crontab (string)
	# return True if period are the same, false if they are different	
	# @author Diana
	def __compareTask (self,t,l):
		if (t == "1 min") and (l[0:8]=="* * * * *"):
			return True
		if (t == "5 min") and (l[0:10]=="*/5 * * * *"):
			return True
		if (t == "10 min") and (l[0:10]=="*/10 * * * *"):
			return True
		if (t == "30 min") and (l[0:10]=="*/30 * * * *"):
			return True
		if (t == "1 hour") and (l[0:10]=="0 */1 * * *"):
			return True
		if (t == "1 day") and ((l[0]!="*") and (l[0]!="0")):
			print l[0]
			return True
