# -*- coding: UTF-8 -*-

#Copyright (C) 2008 Adrián Bravo, Jorge Quintás

#This file is part of HDLorean.
#
#HDLorean is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#HDLorean is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with HDLorean.  If not, see <http://www.gnu.org/licenses/>.


#imports for using pyinotify
from pyinotify import EventsCodes,ProcessEvent,ThreadedNotifier,WatchManager
import inotify
import os.path
from ConfigFileManager import ConfigFileManager
import os
import GlobalExceptions
from hdlogger import HDLogger

## @brief a handler for inotiy events
#
# This module allows HD Lorean to watch events in the directory tree
# We obtain a whitelist and a blacklist of files from the module ConfigFileManager and
# use them to enqueue the events
# @author: Jorge
class PyinotifyHandler:

	
	## Contains paths and masks for watched resources
	__watched = None
	## a threaded notifier of events
	__notifier = None
	# FIXME: __dic is unnecessary for this module functionality, only for testing
	## a dictionary, contains paths and internal descriptors. Paths are the keys
	__pathDic = {}
	##dictionary containing same information but with watch descriptor as key
	__wdDic = {}
	## inhertis from pyinotify.ProcessEvent Contains functions to handle events
	__handler = None
	## list of files/dirs that we don't want to monitorize
	__blacklist = None
	## list of blacklisted patterns
	__reBlacklist = None
	## list of files/dirs that we want to monitorize
	__whitelist = None
	##list of parent directories watched
	__parentlist = None
	## watched list
	__watchedlist = None
	##
	__logger = None
	##
	__log = None
	## contains the number of watches
	__nWatches = 0
	##indicates the status of the notifier
	__isStarted = False


	## @brief takes a list of paths and adds each one to watched.
	#
	# An instance of ConfigFileManager is created to obtain the watched, white and reBlack sets
	# The handler is initialized with those sets
	# a notifier is created with the watched set and the handler
	# @see Watcher
	# @param start if start is true, the notification of events is going to start just after create the
	# instance if pyinotifyhandler
	# @author: Jorge
	def __init__(self,handler,watched,blacklist,whitelist,reBlacklist,parentsList,start):
		self.__logger = HDLogger.HDLogger()
		self.__log = self.__logger.setupHDLogger('hdlorean.Watcher.PyinotifyHandler')
		try:
			self.__watched = WatchManager()
		except GlobalExceptions.TypeMissmatch:
			#TODO: meh, duplicated warnings CFM-Here
			self.__log.error('type missmatch')
		self.__blacklist = blacklist
		self.__whitelist = whitelist
		self.__reBlacklist = reBlacklist
		self.__parentlist = parentsList
		self.__watchedlist = watched
		self.__handler = handler
		if start:
			# expanded files
			for item in watched:
				try:
					self.addToWatched(item)
				except GlobalExceptions.InvalidPath:
					self.__log.error('Invalid path ' + item)
			#parents
			for item in parentsList:
				try:
					self.addParentToWatched(item)
				except GlobalExceptions.InvalidPath:
					self.__log.error('Invalid path ' + item)
			for item in blacklist:
				self.removeFromWatched(item)
			for item in reBlacklist:
				self.removeFromWatched(item)
			self.__notifier = ThreadedNotifier(self.__watched,self.__handler)
	#deprecated
	## @brief Callback method to be notified by Watcher of the changes made in the lists
	#
	# @author Jorge
	def updateLists(self):
		self.__blacklist = blacklist
		self.__whitelist = whitelist
		self.__reBlacklist = reBlacklist
		#TODO: add a try..catch block here
		for add in added:
			try:
				self.addToWatched(add)
				#print 'added '+ add
			except GlobalExceptions.InvalidPath:
				self.__log.error('Invald path' + add)
		for rm in removed:
			self.removeFromWatched(rm)
			#print 'removed ' + rm
		for add in parentAdded:
			self.addParentToWatched(add)
			#print 'added p '+ add
		for rm in parentRemoved:
			#remove from watched can be used to remove parents
			self.removeFromWatched(rm)
			#print 'removed p'+ rm

	## @brief adds a single path to __watched.
	#
	# checks if path exist and if it refers to file or directory
	# then adds path to watched resources with the appropriate flags
	# @param path type=string  add the file or directoriy path to watched
	# @author Jorge
	def addToWatched(self,path):
	# first of all we must know if the path exist, then if it's file or dir
	# TODO: it seems that WatchManager.add checks if the path is valid...check it
		if os.path.exists(path):
			if os.path.isdir(path):
				#mask is provisional
				#TODO put mask as a constant out of here
				mask = EventsCodes.IN_DELETE_SELF | EventsCodes.IN_MODIFY |EventsCodes. IN_CREATE \
					| EventsCodes.IN_DELETE | EventsCodes.IN_MOVED_TO | EventsCodes.IN_ATTRIB | \
					EventsCodes.IN_MOVED_FROM
				_dic = self.__watched.add_watch(path, mask, None, True, True)
				self.__updateDictionaries(_dic,True)
				#self.__log.debug('dictionary paths '+repr(self.__pathDic))
				#self.__log.debug('dictionary wd '+repr(self.__wdDic))
				self.__nWatches = len(self.__pathDic)
				self.__log.debug('Addition. Number of watches = '+ repr(self.__nWatches))
				if(self.__nWatches >= 0.8 * inotify.max_user_watches.value):
					self.__log.warning('Number of watches is reaching the limit')
			else:
				#mask is provisional
				#mask = EventsCodes.IN_CREATE | EventsCodes.IN_MODIFY | EventsCodes.IN_DELETE_SELF | \
				#		EventsCodes.IN_ATTRIB | EventsCodes.IN_MOVED_TO | EventsCodes.IN_MOVED_FROM
				#self.__watched.add_watch(path, mask, None)
				#We're not going to directly watch files, only directories
				#TODO: refactor this if-else
					pass
		else:
			raise GlobalExceptions.InvalidPath
		

	## @brief add a directorie to __watched, non recursively
	#
	# @param path the path of the parent directory  
	def addParentToWatched(self,path):
		if os.path.isdir(path):
			mask = EventsCodes.IN_DELETE_SELF | EventsCodes.IN_MODIFY |EventsCodes. IN_CREATE \
				| EventsCodes.IN_DELETE | EventsCodes.IN_MOVE_SELF | EventsCodes.IN_MOVED_TO | \
				 EventsCodes.IN_MOVED_FROM
			_dic = self.__watched.add_watch(path, mask, self.__handler, False, True)
			#self.__updateDictionaries(_dic,True)
			self.__nWatches = len(self.__pathDic)
			self.__log.debug('Addition. Number of watches = ' + repr(self.__nWatches))
			if(self.__nWatches >= 0.8 * inotify.max_user_watches.value):
				self.__log.warning('Number of watches is reaching the limit')
		else:
			raise GlobalExceptions.InvalidPath
	
	#deprecated. Used to watch config with inotify
	def addConfigToWatched(self):
		mask = EventsCodes.IN_DELETE_SELF | EventsCodes.IN_MODIFY | EventsCodes.IN_DELETE
		path= os.path.join(os.path.join(os.getenv('HOME'),'.hdlorean'),'config.cfg')
		self.__dic = self.__watched.add_watch(path,mask,None)
		self.__nWatches = len(self.__pathDic)
		self.__log.debug('Addition. Number of watches = '+ repr(self.__nWatches))
		if(self.__nWatches >= 0.8 * inotify.max_user_watches.value):
			self.__log.warning('Number of watches is reaching the limit')
		#mask = EventsCodes.IN_CREATE
		#path = os.path.join(os.getenv('HOME'),'.hdlorean')
		#self.__watched.add_watch(path, mask,None)
		#self.__log.info('config is being watched')

	

	## @brief Run the notifier in a separate thread.
	#
	# @author: Jorge
	def startNotifier(self):
		if not(self.__notifier.isAlive()):
			self.__notifier.start()
			self.__isStarted = True
		else:
			raise GlobalExceptions.NotifierRunning
	
	## @brief stop the notifier
	#
	# @author: Jorge
	def stopNotifier(self):
		if self.__notifier.isAlive():
			self.__notifier.stop()
			self.__isStarted = False
		else:
			raise GlobalExceptions.NotifierStopped
			
			
	def removeFromWatched(self,path):
	#pyinotify don't return a dictionary of wd and paths, instead of this we obtain
	#a dictionary of wd and boolean
		if path != '':
			#wd = self.__watched.get_wd(path)  <-- this is soooo slow
			try:
				wd = self.__pathDic[path]
			except KeyError:
				wd = -1
			#the second parameter is True, watches are deleted recursively
			_dic = self.__watched.rm_watch(wd,True)
			#here is the reconstruction of wd-path dictionary needed
			#deleteMe will contain the keys that must be deleted from _dic. _dic contains watch descriptors of the
			#files and directories that have been modified while path was watched. The wd of these files/directories is not
			#stored in wdDic nor pathDic, so if we keep this information into _dic we will fail  trying to _updateDictionaries
			deleteMe = []
			for key in _dic:
				if _dic[key]:
					# si se han creado cosas en un directorio que estamos borrando, _dic va a dar claves que no están en wdDic
					# si estas claves son true, no pasa nada, se ha quitado el watch. Si son falsas, no se ha podido quitar y hay que 
					# tratarlas
					try:
						_dic[key] = self.__wdDic[key]
					except KeyError:
						#borramos la entrada que da error en _wdDic
						deleteMe.append(key)
				else:
					deleteMe.append(key)
			#here we proceed with the deletion of the pairs that we don't have on __wdDic nor __pathDic
			for rm in deleteMe:
				del _dic[rm]
			self.__updateDictionaries(_dic,False)
			self.__nWatches = len(self.__pathDic)
			self.__log.debug('Remove. Number of watches =' + repr(self.__nWatches))
	
	## @brief Updates the internal list and watched
	#
	# to update watched we receive from CFM two **lists** containing what has
	# disappeared from watched, and what has appeared. With this info we add and remove
	# watches
	# @param blacklist the new blacklist
	# @param whitelist the new whitelist
	# @param reBlacklist the new pattern blacklist
	# @param added contains the paths of files/dirs that must be added to wathced 
	# @param removed like added but reffers to remove
	# @author: Jorge
	#FIXME: this function and removeFromWatched are WIP...WIP!!!!
	def updateWatched(self,blacklist,whitelist,reBlacklist,added,removed,parentAdded,parentRemoved):
		self.__blacklist = blacklist
		self.__whitelist = whitelist
		self.__reBlacklist = reBlacklist
		#add and remove in watched
		for add in added:
			try:
				self.addToWatched(add)
				self.__watchedlist.add(add)
			except GlobalExceptions.InvalidPath:
				self.__log.error('Invald path' + add)
		for rm in removed:
			self.removeFromWatched(rm)
			self.__watchedlist.discard(rm)
		#same as above with parents
		for add in parentAdded:
			self.addParentToWatched(add)
			#update local list
			self.__parentlist.add(add)
		for rm in parentRemoved:
			#remove from watched can be used to remove parents
			self.removeFromWatched(rm)
			self.__parentlist.discard(rm)
	
	##@brief returns the parentslist
	#
	def getParentList(self):
		return self.__parentlist
	
	##@brief updates internal dictionaries when watches are added or removed
	#
	#@param dictionary contains the elements to add/remove
	#@param add selects the operation mode, add or remove from the internal dictionaries 
	def __updateDictionaries(self,dictionary,add):
		if add:
			self.__pathDic.update(dictionary)
			#print 'dictionary ----------> '+repr(dictionary)
			#print 'new------------------> '+repr(self.__pathDic)
			#the next is more simple than it appears
			# the wdDic keys are the values of dictionary
			for key in dictionary:
				self.__wdDic[dictionary[key]] = key
		else:
			for key in dictionary:
				del self.__pathDic[dictionary[key]]
				del self.__wdDic[key]
				
				
	def stop(self):
		for path in self.__watchedlist:
			self.removeFromWatched(path)
		for path in self.__parentlist:
			self.removeFromWatched(path)
			
			
	def start(self):		
		for path in self.__watchedlist:
			self.addToWatched(path)
		for path in self.__parentlist:
			self.addToWatched(path)
		if not self.__isStarted:
			for item in self.__blacklist:
				self.removeFromWatched(item)
			for item in self.__reBlacklist:
				self.removeFromWatched(item)
			self.__notifier = ThreadedNotifier(self.__watched,self.__handler)			
			self.startNotifier()
			self.__log.info('Started')
					
			
	def update(self,mask):
		if mask == False:
			mask = None
		else:
			mask = EventsCodes.IN_DELETE_SELF | EventsCodes.IN_MODIFY |EventsCodes. IN_CREATE \
			| EventsCodes.IN_DELETE | EventsCodes.IN_MOVED_TO | EventsCodes.IN_ATTRIB | \
			EventsCodes.IN_MOVED_FROM
		for item in self.__watchedlist: 
			wd = self.__pathDic[item]
			self.__watched.update_watch(wd, mask,None, True)
##
##		FOR DEBUGING PROPOSAL
##REMEMBER: This CODE has been moved into Watcher

#ph = PyinotifyHandler()
#print 'Succesfully initialized'
#ph.startNotifier()
#print 'Started'
#while True:
#	try:
#		import time
#		time.sleep(5)
#	except KeyboardInterrupt:
#		# ...until c^c signal
#		print 'stop monitoring...'
#		# stop monitoring
#		ph.stopNotifier()
#		break
#	except Exception, err:
#		# otherwise keep on looping
#		print err
